from __future__ import annotations

import time

__all__ = ["Adapter"]

from unicex.types import (
    KlineDict,
    OpenInterestDict,
    OpenInterestItem,
    TickerDailyDict,
    TickerDailyItem,
)
from unicex.utils import catch_adapter_errors, decorate_all_methods


@decorate_all_methods(catch_adapter_errors)
class Adapter:
    """Адаптер для унификации данных с Gateio API."""

    @staticmethod
    def tickers(raw_data: list[dict], only_usdt: bool) -> list[str]:
        """Преобразует сырой ответ о тикерах в список символов.

        Параметры:
            raw_data (list[dict]): Сырой ответ с биржи.
            only_usdt (bool): Флаг, указывающий, нужно ли включать только тикеры в паре c USDT.

        Возвращает:
            list[str]: Список тикеров.
        """
        return [
            item["currency_pair"]
            for item in raw_data
            if item["currency_pair"].endswith("USDT") or not only_usdt
        ]

    @staticmethod
    def futures_tickers(raw_data: list[dict], only_usdt: bool) -> list[str]:
        """Преобразует сырой ответ о фьючерсных тикерах в список символов.

        Параметры:
            raw_data (list[dict]): Сырой ответ с биржи.
            only_usdt (bool): Флаг, указывающий, нужно ли включать только тикеры в паре c USDT.

        Возвращает:
            list[str]: Список тикеров.
        """
        return [
            item["contract"]
            for item in raw_data
            if item["contract"].endswith("USDT") or not only_usdt
        ]

    @staticmethod
    def last_price(raw_data: list[dict]) -> dict[str, float]:
        """Преобразует данные о последних ценах (spot) в унифицированный формат.

        Параметры:
            raw_data (list[dict]): Сырой ответ с биржи.

        Возвращает:
            dict[str, float]: Словарь, где ключ — тикер, а значение — последняя цена.
        """
        return {item["currency_pair"]: float(item["last"]) for item in raw_data}

    @staticmethod
    def futures_last_price(raw_data: list[dict]) -> dict[str, float]:
        """Преобразует данные о последних ценах (futures) в унифицированный формат."""
        return {item["contract"]: float(item["last"]) for item in raw_data}

    @staticmethod
    def ticker_24hr(raw_data: list[dict]) -> TickerDailyDict:
        """Преобразует 24-часовую статистику (spot) в унифицированный формат.

        Параметры:
            raw_data (list[dict]): Сырой ответ с биржи.

        Возвращает:
            TickerDailyDict: Словарь, где ключ — тикер, а значение — агрегированная статистика.
        """
        return {
            item["currency_pair"]: TickerDailyItem(
                p=float(item["change_percentage"]),
                v=float(item["base_volume"]),
                q=float(item["quote_volume"]),
            )
            for item in raw_data
        }

    @staticmethod
    def futures_ticker_24hr(raw_data: list[dict]) -> TickerDailyDict:
        """Преобразует 24-часовую статистику (futures) в унифицированный формат."""
        return {
            item["contract"]: TickerDailyItem(
                p=float(item["change_percentage"]),
                v=float(item["volume_24h_base"]),
                q=float(item["volume_24h_quote"]),
            )
            for item in raw_data
        }

    @staticmethod
    def klines(raw_data: list[list], symbol: str) -> list[KlineDict]:
        """Преобразует данные о свечах в унифицированный формат.

        Параметры:
            raw_data (list[list]): Сырой ответ с биржи.
            symbol (str): Символ тикера.

        Возвращает:
            list[KlineDict]: Список свечей.
        """
        return [
            KlineDict(
                s=symbol,
                t=int(kline[0]),
                o=float(kline[5]),
                h=float(kline[3]),
                l=float(kline[4]),
                c=float(kline[2]),
                v=float(kline[6]),
                q=float(kline[1]),
                T=None,
                x=kline[7] == "true",
            )
            for kline in sorted(
                raw_data,
                key=lambda x: int(x[0]),
            )
        ]

    @staticmethod
    def futures_klines(raw_data: list[dict], symbol: str) -> list[KlineDict]:
        """Преобразует данные о свечах в унифицированный формат.

        Параметры:
            raw_data (list[dict]): Сырой ответ с биржи.
            symbol (str): Символ тикера.

        Возвращает:
            list[KlineDict]: Список свечей.
        """
        return [
            KlineDict(
                s=symbol,
                t=int(kline["t"]) * 1000,  # переводим секунды → миллисекунды
                o=float(kline["o"]),
                h=float(kline["h"]),
                l=float(kline["l"]),
                c=float(kline["c"]),
                v=float(kline["v"]),
                q=float(kline["sum"]),  # "sum" = объем в $ (quote volume)
                T=None,
                x=None,
            )
            for kline in sorted(raw_data, key=lambda x: int(x["t"]))
        ]

    @staticmethod
    def funding_rate(raw_data: list[dict]) -> dict[str, float]:
        """Преобразует данные о ставках финансирования в унифицированный формат."""
        return {
            item["contract"]: float(item["funding_rate"]) * 100
            for item in raw_data
            if item.get("funding_rate") is not None
        }

    @staticmethod
    def open_interest(raw_data: list[dict]) -> OpenInterestDict:
        """Преобразует данные об открытом интересе в унифицированный формат."""
        return {
            item["contract"]: OpenInterestItem(
                t=int(time.time() * 1000),
                v=float(item["total_size"]) * float(item["quanto_multiplier"]),
            )
            for item in raw_data
        }



__all__ = ["Adapter"]


from unicex.types import (
    AggTradeDict,
    KlineDict,
    OpenInterestItem,
    TickerDailyDict,
    TickerDailyItem,
    TradeDict,
)
from unicex.utils import catch_adapter_errors, decorate_all_methods


@decorate_all_methods(catch_adapter_errors)
class Adapter:
    """Адаптер для унификации данных с Binance API."""

    @staticmethod
    def tickers(raw_data: list[dict], only_usdt: bool) -> list[str]:
        """Преобразует сырой ответ, в котором содержатся данные о тикерах в список тикеров.

        Параметры:
            raw_data (Any): Сырой ответ с биржи.
            only_usdt (bool): Флаг, указывающий, нужно ли включать только тикеры в паре к USDT.

        Возвращает:
            list[str]: Список тикеров.
        """
        return [
            item["symbol"] for item in raw_data if item["symbol"].endswith("USDT") or not only_usdt
        ]

    @staticmethod
    def ticker_24hr(raw_data: list[dict]) -> TickerDailyDict:
        """Преобразует сырой ответ, в котором содержатся данные о тикере за последние 24 часа в унифицированный формат.

        Параметры:
            raw_data (Any): Сырой ответ с биржи.

        Возвращает:
            TickerDailyDict: Словарь, где ключ - тикер, а значение - статистика за последние 24 часа.
        """
        return {
            item["symbol"]: TickerDailyItem(
                p=float(item["priceChangePercent"]),
                q=float(item["quoteVolume"]),  # объём в долларах
                v=float(item["volume"]),  # объём в монетах
            )
            for item in raw_data
        }

    @staticmethod
    def last_price(raw_data: list[dict]) -> dict[str, float]:
        """Преобразует сырой ответ, в котором содержатся данные о тикере за последние 24 часа в унифицированный формат.

        Параметры:
            raw_data (list[dict]): Сырой ответ с биржи.

        Возвращает:
            dict[str, float]: Словарь, где ключ - тикер, а значение - последняя цена.
        """
        return {item["symbol"]: float(item["price"]) for item in raw_data}

    @staticmethod
    def futures_last_price(raw_data: list[dict]) -> dict[str, float]:
        """Преобразует сырой ответ, в котором содержатся данные о тикере за последние 24 часа в унифицированный формат.

        Параметры:
            raw_data (list[dict]): Сырой ответ с биржи.

        Возвращает:
            dict[str, float]: Словарь, где ключ - тикер, а значение - последняя цена.
        """
        return Adapter.last_price(raw_data)

    @staticmethod
    def klines(raw_data: list[list], symbol: str) -> list[KlineDict]:
        """Преобразует сырой ответ, в котором содержатся данные о котировках тикеров в унифицированный формат.

        Параметры:
            raw_data (list[list]): Сырой ответ с биржи.
            symbol (str): Символ тикера.

        Возвращает:
            list[KlineDict]: Список словарей, где каждый словарь содержит данные о свече.
        """
        return [
            KlineDict(
                s=symbol,
                t=kline[0],
                o=float(kline[1]),
                h=float(kline[2]),
                l=float(kline[3]),
                c=float(kline[4]),
                v=float(kline[5]),
                q=float(kline[7]),
                T=kline[6],
                x=None,
            )
            for kline in sorted(
                raw_data,
                key=lambda x: int(x[0]),
            )
        ]

    @staticmethod
    def funding_rate(raw_data: list[dict]) -> dict[str, float]:
        """Преобразует сырой ответ, в котором содержатся данные о ставках финансирования тикеров в унифицированный формат.

        Параметры:
            raw_data (list[dict]): Сырой ответ с биржи.

        Возвращает:
            dict[str, float]: Словарь, где ключ - тикер, а значение - ставка финансирования.
        """
        return {item["symbol"]: float(item["lastFundingRate"]) * 100 for item in raw_data}

    @staticmethod
    def open_interest(raw_data: dict) -> OpenInterestItem:
        """Преобразует сырое сообщение с вебсокета, в котором содержится информация о
        объеме открытых позиций в унифицированный вид.

        Параметры:
            raw_data (Any): Сырое сообщение с вебсокета.

        Возвращает:
            OpenInterestItem: Словарь со временем и объемом открытого интереса в монетах.
        """
        return OpenInterestItem(t=raw_data["time"], v=float(raw_data["openInterest"]))

    @staticmethod
    def klines_message(raw_msg: dict) -> list[KlineDict]:
        """Преобразует сырое сообщение с вебсокета, в котором содержится информация о
        свече/свечах в унифицированный вид.

        Параметры:
            raw_msg (Any): Сырое сообщение с вебсокета.

        Возвращает:
            list[KlineDict]: Список словарей, где каждый словарь содержит данные о свече.
        """
        # Обрабатываем обертку в случае с multiplex stream
        kline = raw_msg.get("data", raw_msg)["k"]
        return [
            KlineDict(
                s=kline["s"],
                t=kline["t"],
                o=float(kline["o"]),
                h=float(kline["h"]),
                l=float(kline["l"]),
                c=float(kline["c"]),
                v=float(kline["v"]),  # Используем quote volume (в USDT)
                q=float(kline["q"]),  # Используем quote volume (в USDT)
                T=kline["T"],
                x=kline["x"],
            )
        ]

    @staticmethod
    def aggtrades_message(raw_msg: dict) -> list[AggTradeDict]:
        """Преобразует сырое сообщение с вебсокета, в котором содержится информация о
        аггрегированных сделке/сделках в унифицированный вид.

        Параметры:
            raw_msg (Any): Сырое сообщение с вебсокета.

        Возвращает:
            list[KlineDict]: Список словарей, где каждый словарь содержит данные о сделке.
        """
        msg = raw_msg.get("data", raw_msg)
        return [
            AggTradeDict(
                t=int(msg["T"]),
                s=str(msg["s"]),
                S="SELL" if bool(msg["m"]) else "BUY",
                p=float(msg["p"]),
                v=float(msg["q"]),
            )
        ]

    @staticmethod
    def trades_message(raw_msg: dict) -> list[TradeDict]:
        """Преобразует сырое сообщение с вебсокета, в котором содержится информация о
        сделке/сделках в унифицированный вид.

        Параметры:
            raw_msg (Any): Сырое сообщение с вебсокета.

        Возвращает:
            list[KlineDict]: Список словарей, где каждый словарь содержит данные о сделке.
        """
        msg = raw_msg.get("data", raw_msg)
        return [
            TradeDict(
                t=int(msg["T"]),
                s=str(msg["s"]),
                S="SELL" if bool(msg["m"]) else "BUY",
                p=float(msg["p"]),
                v=float(msg["q"]),
            )
        ]



"""Модуль, который предоставляет дополнительные функции, которые могут пригодиться в работе."""

__all__ = [
    "percent_greater",
    "percent_less",
    "TimeoutTracker",
    "generate_ex_link",
    "generate_tv_link",
    "generate_cg_link",
    "make_humanreadable",
    "normalize_ticker",
    "normalize_symbol",
]

import time
from typing import Literal

from .enums import Exchange, MarketType
from .exceptions import NotSupported


def percent_greater(higher: float, lower: float) -> float:
    """Возвращает на сколько процентов `higher` больше `lower`.

    Можно воспринимать полученное значение как если вести линейку на tradingview.com от меньшего значения к большему.

    Например:
        ```python
        percent_greater(120, 100)
        >> 20.0
        ```

    Возвращает:
        `float`: На сколько процентов `higher` больше `lower`.
    """
    if lower == 0:
        return 0.0  # Не будем возвращать float('inf'), чтобы не ломать логику приложения
    return (higher / lower - 1) * 100


def percent_less(higher: float, lower: float) -> float:
    """Возвращает на сколько процентов `lower` меньше `higher`.

    Можно воспринимать полученное значение как если вести линейку на tradingview.com от большего значения к меньшему.

    Например:
        ```python
        percent_less(120, 100)
        >> 16.67777777777777
        ```

    Возвращает:
        `float`: На сколько процентов `lower` меньше `higher`.
    """
    if lower == 0:
        return 0.0  # Не будем возвращать float('inf'), чтобы не ломать логику приложения
    return (1 - lower / higher) * 100


class TimeoutTracker[T]:
    """Универсальный менеджер для управления таймаутами любых объектов.
    Позволяет временно блокировать объекты на заданный промежуток времени.
    """

    def __init__(self) -> None:
        """Инициализирует пустой словарь для отслеживания заблокированных объектов."""
        self._blocked_items: dict[T, float] = {}

    def is_blocked(self, item: T) -> bool:
        """Проверяет, находится ли объект в состоянии блокировки.
        Если срок блокировки истёк, удаляет объект из списка.

        Параметры:
            item (`T`): Объект, который нужно проверить.

        Возвращает:
            `bool`: True, если объект заблокирован, иначе False.
        """
        if item in self._blocked_items:
            if time.time() < self._blocked_items[item]:
                return True
            else:
                del self._blocked_items[item]
        return False

    def block(self, item: T, duration: int) -> None:
        """Блокирует объект на указанное количество секунд.

        Параметры:
            item (`T`): Объект, который нужно заблокировать.
            duration (`int`): Длительность блокировки в секундах.
        """
        self._blocked_items[item] = time.time() + duration


def normalize_ticker(raw_ticker: str) -> str:
    """Нормализует тикер и возвращает базовую валюту (например, `BTC`).

    Эта функция принимает тикер в различных форматах (с разделителями, постфиксом SWAP,
    в верхнем или нижнем регистре) и приводит его к стандартному виду — только базовый актив.

    Примеры:
        ```python
        normalize_ticker("BTC-USDT")  # "BTC"
        normalize_ticker("BTC-USDT-SWAP")  # "BTC"
        normalize_ticker("btc_usdt")  # "BTC"
        normalize_ticker("BTCUSDT")  # "BTC"
        normalize_ticker("BTC")  # "BTC"
        ```

    Параметры:
        raw_ticker (`str`): Исходный тикер в любом из распространённых форматов.

    Возвращает:
        `str`: Базовый актив в верхнем регистре (например, `"BTC"`).
    """
    ticker = raw_ticker.upper()

    # Удаляем постфиксы SWAP
    if ticker.endswith(("SWAP", "-SWAP", "_SWAP", ".SWAP")):
        ticker = (
            ticker.removesuffix("-SWAP")
            .removesuffix("_SWAP")
            .removesuffix(".SWAP")
            .removesuffix("SWAP")
        )

    # Удаляем разделители
    ticker = ticker.translate(str.maketrans("", "", "-_."))

    # Убираем суффикс валюты котировки
    for quote in ("USDT", "USDC"):
        if ticker.endswith(quote):
            ticker = ticker.removesuffix(quote)
            break

    return ticker


def normalize_symbol(raw_ticker: str, quote: Literal["USDT", "USDC"] = "USDT") -> str:
    """Нормализует тикер до унифицированного символа (например, `BTCUSDT`).

    Функция принимает тикер в любом из популярных форматов и возвращает полный символ,
    состоящий из базовой валюты и указанной валюты котировки (`USDT` или `USDC`).

    Примеры:
        ```python
        normalize_symbol("BTC-USDT")  # "BTCUSDT"
        normalize_symbol("BTC")  # "BTCUSDT"
        normalize_symbol("btc_usdt_swap")  # "BTCUSDT"
        normalize_symbol("ETH", "USDC")  # "ETHUSDC"
        ```

    Параметры:
        raw_ticker (`str`): Исходный тикер в любом из распространённых форматов.
        quote (`Literal["USDT", "USDC"]`, optional): Валюта котировки.
            По умолчанию `"USDT"`.

    Возвращает:
        `str`: Символ в унифицированном формате, например `"BTCUSDT"`.
    """
    base = normalize_ticker(raw_ticker)
    return f"{base}{quote}"


def generate_ex_link(exchange: Exchange, market_type: MarketType, symbol: str):
    """Генерирует ссылку на биржу.

    Параметры:
        exchange (`Exchange`): Биржа.
        market_type (`MarketType`): Тип рынка.
        symbol (`str`): Символ.

    Возвращает:
        `str`: Ссылка на биржу.
    """
    symbol = normalize_symbol(symbol)
    ticker = normalize_ticker(symbol)
    if exchange == Exchange.BINANCE:
        if market_type == MarketType.FUTURES:
            return f"https://www.binance.com/en/futures/{symbol}"
        else:
            return f"https://www.binance.com/en/trade/{ticker}_USDT?type=spot"
    elif exchange == Exchange.BYBIT:
        if market_type == MarketType.FUTURES:
            return f"https://www.bybit.com/trade/usdt/{symbol}"
        else:
            return f"https://www.bybit.com/en/trade/spot/{ticker}/USDT"
    elif exchange == Exchange.BITGET:
        if market_type == MarketType.FUTURES:
            return f"https://www.bitget.com/ru/futures/usdt/{symbol}"
        else:
            return f"https://www.bitget.com/ru/spot/{symbol}"
    elif exchange == Exchange.OKX:
        if market_type == MarketType.FUTURES:
            return f"https://www.okx.com/ru/trade-swap/{ticker.lower()}-usdt-swap"
        else:
            return f"https://www.okx.com/ru/trade-spot/{ticker.lower()}-usdt"
    elif exchange == Exchange.MEXC:
        if market_type == MarketType.FUTURES:
            return f"https://www.mexc.com/ru-RU/futures/{ticker}_USDT?type=linear_swap"
        else:
            return f"https://www.mexc.com/ru-RU/exchange/{ticker}_USDT"
    elif exchange == Exchange.GATE:
        if market_type == MarketType.FUTURES:
            return f"https://www.gate.com/ru/futures/USDT/{ticker}_USDT"
        else:
            return f"https://www.gate.com/ru/trade/{ticker}_USDT"
    elif exchange == Exchange.XT:
        if market_type == MarketType.FUTURES:
            return f"https://www.xt.com/ru/futures/trade/{ticker.lower()}_usdt"
        else:
            return f"https://www.xt.com/ru/trade/{ticker.lower()}_usdt"
    elif exchange == Exchange.BITUNIX:
        if market_type == MarketType.FUTURES:
            return f"https://www.bitunix.com/ru-ru/contract-trade/{ticker.upper()}USDT"
        else:
            return f"https://www.bitunix.com/ru-ru/spot-trade/{ticker.upper()}USDT"
    elif exchange == Exchange.KCEX:
        if market_type == MarketType.FUTURES:
            return f"https://www.kcex.com/ru-RU/futures/exchange/{ticker.upper()}_USDT"
        else:
            return f"https://www.kcex.com/ru-RU/exchange/{ticker.upper()}_USDT"
    elif exchange == Exchange.HYPERLIQUID:
        if market_type == MarketType.FUTURES:
            return f"https://app.hyperliquid.xyz/trade/{ticker}"
        else:
            return f"https://app.hyperliquid.xyz/trade/{ticker}/USDC"
    else:
        raise NotSupported(f"Exchange {exchange} is not supported")


def generate_tv_link(exchange: Exchange, market_type: MarketType, symbol: str) -> str:
    """Генерирует ссылку для TradingView.

    Параметры:
        exchange (`Exchange`): Биржа.
        market_type (`MarketType`): Тип рынка.
        symbol (`str`): Символ.

    Возвращает:
        `str`: Ссылка для TradingView.
    """
    symbol = normalize_symbol(symbol)
    exchange_str = "GATEIO" if exchange == Exchange.GATE else str(exchange)
    if market_type == MarketType.FUTURES:
        return f"https://www.tradingview.com/chart/?symbol={exchange_str}:{symbol}.P"
    else:
        return f"https://www.tradingview.com/chart/?symbol={exchange_str}:{symbol}"


def generate_cg_link(exchange: Exchange, market_type: MarketType, symbol: str) -> str:
    """Генерирует ссылку для CoinGlass.

    Параметры:
        exchange (`Exchange`): Биржа.
        market_type (`MarketType`): Тип рынка.
        symbol (`str`): Символ.

    Возвращает:
        `str`: Ссылка для CoinGlass.
    """
    base_url = "https://www.coinglass.com/tv/ru"

    symbol = normalize_symbol(symbol)

    if market_type == MarketType.FUTURES:
        match exchange:
            case Exchange.OKX:
                return f"{base_url}/OKX_{symbol.replace('USDT', '-USDT')}-SWAP"
            case Exchange.MEXC:
                return f"{base_url}/MEXC_{symbol.replace('USDT', '_USDT')}"
            case Exchange.BITGET:
                return f"{base_url}/Bitget_{symbol}_UMCBL"
            case Exchange.GATE:
                return f"{base_url}/Gate_{symbol.replace('USDT', '_USDT')}"
            case Exchange.BITUNIX:
                return f"{base_url}/Bitunix_{symbol}"
            case Exchange.HYPERLIQUID:
                return f"{base_url}/Hyperliquid_{symbol.replace('USDT', '-USD')}"
            case _:
                return f"{base_url}/{exchange.capitalize()}_{symbol}"
    else:
        # Для спота корректная ссылка есть только у OKX
        if exchange == Exchange.OKX:
            return f"{base_url}/SPOT_{exchange.upper()}_{symbol.replace('USDT', '-USDT')}"
        # Для остальных бирж ссылки нет → возвращаем заглушку
        return generate_cg_link(exchange, MarketType.FUTURES, symbol)


def make_humanreadable(value: float, locale: Literal["ru", "en"] = "ru") -> str:
    """Функция превращает большие числа в удобочитаемый вид.

    Принимает:
        value (float): число для преобразования
        locale (Literal["ru", "en"]): язык для форматирования числа

    Возвращает:
        str: Человеческое представление числа
    """
    suffixes = {
        "ru": {
            1_000: "тыс.",
            1_000_000: "млн.",
            1_000_000_000: "млрд.",
            1_000_000_000_000: "трлн.",
            1_000_000_000_000_000: "квдрлн.",
            1_000_000_000_000_000_000: "квнтлн.",
        },
        "en": {
            1_000: "K",
            1_000_000: "M",
            1_000_000_000: "B",
            1_000_000_000_000: "T",
            1_000_000_000_000_000: "Qa",  # Quadrillion
            1_000_000_000_000_000_000: "Qi",  # Quintillion
        },
    }

    selected_suffixes = suffixes[locale]

    for divisor in sorted(selected_suffixes.keys(), reverse=True):
        if abs(value) >= divisor:
            number = value / divisor
            if locale == "ru":
                return (
                    f"{number:,.2f}".replace(",", " ").replace(".", ",")
                    + f" {selected_suffixes[divisor]}"
                )
            return f"{number:,.2f} {selected_suffixes[divisor]}"

    # Форматирование "малых" чисел
    if locale == "ru":
        return f"{value:,.2f}".replace(",", " ").replace(".", ",")
    return f"{value:,.2f}"



__all__ = ["ExchangeInfo"]

import aiohttp

from unicex._abc import IExchangeInfo
from unicex.types import TickerInfoItem

from .client import Client


class ExchangeInfo(IExchangeInfo):
    """Предзагружает информацию о тикерах для биржи Okx."""

    exchange_name = "Okx"
    """Название биржи, на которой работает класс."""

    @classmethod
    async def _load_spot_exchange_info(cls, session: aiohttp.ClientSession) -> None:
        """Загружает информацию о бирже для спотового рынка."""
        tickers_info = {}
        exchange_info = await Client(session).get_instruments("SPOT")
        for el in exchange_info["data"]:
            tickers_info[el["instId"]] = TickerInfoItem(
                tick_precision=None,
                tick_step=float(el["tickSz"]),
                size_precision=None,
                size_step=float(el["lotSz"]),
                contract_size=1,
            )

        cls._tickers_info = tickers_info

    @classmethod
    async def _load_futures_exchange_info(cls, session: aiohttp.ClientSession) -> None:
        """Загружает информацию о бирже для фьючерсного рынка."""
        tickers_info = {}
        exchange_info = await Client(session).get_instruments("SWAP")
        for el in exchange_info["data"]:
            tickers_info[el["instId"]] = TickerInfoItem(
                tick_precision=None,
                tick_step=float(el["tickSz"]),
                size_precision=None,
                size_step=float(el["lotSz"]) * float(el["ctVal"]),
                contract_size=float(el["ctVal"]),
            )

        cls._futures_tickers_info = tickers_info



__all__ = ["ExchangeInfo"]

import aiohttp

from unicex._abc import IExchangeInfo
from unicex.types import TickerInfoItem

from .client import Client


class ExchangeInfo(IExchangeInfo):
    """Предзагружает информацию о тикерах для биржи Hyperliquid."""

    exchange_name = "Hyperliquid"
    """Название биржи, на которой работает класс."""

    _spot_meta: dict = {}
    """Словарь с метаинформацией о спотовом рынке."""

    _spot_ident_to_idx: dict = {}
    """Словарь, в котором ключ - индетефикатор тикера на бирже, например '@123', а значение - его индекс в _spot_meta."""

    _spot_idx_to_name: dict = {}
    """Словарь, в котором ключ - индекс в _spot_meta, например "123", а значение - название тикера, например 'BTC'."""

    _futures_meta: dict = {}
    """Словарь с метаинформацией о фьючерсном рынке."""

    # DOCS: https://hyperliquid.gitbook.io/hyperliquid-docs/for-developers/api/tick-and-lot-size

    @classmethod
    async def _load_spot_exchange_info(cls, session: aiohttp.ClientSession) -> None:
        """Загружает информацию о бирже для спотового рынка."""
        cls._spot_meta = await Client(session).spot_metadata()
        cls._build_spot_mappings(cls._spot_meta)

        tickers_info: dict[str, TickerInfoItem] = {}
        for symbol_info in cls._spot_meta["tokens"]:
            tickers_info[symbol_info["name"]] = TickerInfoItem(
                tick_step=None,
                tick_precision=int(symbol_info["weiDecimals"]),
                size_step=None,
                size_precision=int(symbol_info["szDecimals"]),
                contract_size=1,
            )

        cls._tickers_info = tickers_info

    @classmethod
    async def _load_futures_exchange_info(cls, session: aiohttp.ClientSession) -> None:
        """Загружает информацию о бирже для фьючерсного рынка."""
        cls._futures_meta = await Client(session).perp_metadata()

    @classmethod
    def _build_spot_mappings(cls, spot_meta: dict) -> None:
        """Строит словари соответствия '@индекс' ↔ индекс ↔ 'BTC'."""
        universe = spot_meta["universe"]
        tokens = spot_meta["tokens"]

        number_to_idx = {}
        for u in universe:
            number_to_idx[u["name"]] = u["tokens"][0]
        cls._spot_ident_to_idx = number_to_idx

        idx_to_name = {}
        for t in tokens:
            idx_to_name[t["index"]] = t["name"]
        cls._spot_idx_to_name = idx_to_name

    @classmethod
    def get_spot_meta(cls) -> dict:
        """Возвращает метаинформацию о спотовом рынке."""
        cls._check_loaded()
        return cls._spot_meta

    @classmethod
    def get_futures_meta(cls) -> dict:
        """Возвращает метаинформацию о фьючерсном рынке."""
        cls._check_loaded()
        return cls._futures_meta

    @classmethod
    def resolve_spot_symbol(cls, ident: str) -> str:
        """Преобразует внутренний идентификатор вида '@142' в тикер, например 'BTC'.
        Не рейзит KeyError, если тикер не найден.

        Параметры:
            token_name (str): Имя токена на бирже, например '@142' или 'BTC'.

        Возвращает:
            str | None: Название тикера (например 'BTC'), либо None, если не найден.
        """
        cls._check_loaded()

        try:
            return cls._spot_idx_to_name[cls._spot_ident_to_idx[ident]]
        except KeyError:
            return ident

    @classmethod
    def resolve_spot_ident(cls, symbol: str) -> str:
        """Преобразует тикер (например, 'BTC') в внутренний идентификатор (например, '@142').

        Параметры:
            symbol (str): Название тикера, например 'BTC'.

        Возвращает:
            str: Внутренний идентификатор (например '@142').

        Исключения:
            KeyError: Если тикер не найден в локальном кэше биржи.
        """
        cls._check_loaded()

        # Находим индекс по тикеру
        idx = next(k for k, v in cls._spot_idx_to_name.items() if v == symbol)

        # Возвращаем внутренний идентификатор вида "@142"
        return next(k for k, v in cls._spot_ident_to_idx.items() if v == idx)



__all__ = ["Client"]


import time
from typing import Any

from unicex._base import BaseClient
from unicex.exceptions import NotAuthorized
from unicex.types import RequestMethod
from unicex.utils import dict_to_query_string, filter_params, generate_hmac_sha256_signature


class Client(BaseClient):
    """Клиент для работы с MEXC Spot API."""

    _BASE_SPOT_URL: str = "https://api.mexc.com"
    """Базовый URL для REST API MEXC."""

    _BASE_FUTURES_URL: str = "https://contract.mexc.com"
    """Базовый URL для фьючерсного REST API MEXC."""

    _RECV_WINDOW: str = "5000"
    """Стандартный интервал времени для получения ответа от сервера."""

    def _get_headers(self, signed: bool = False) -> dict:
        """Формирует заголовки запроса."""
        headers = {"Content-Type": "application/json"}
        if signed:
            if not self._api_key:
                raise NotAuthorized("API key is required for private endpoints.")
            headers["X-MEXC-APIKEY"] = self._api_key
        return headers

    def _generate_signature(self, payload: dict) -> str:
        """Генерирует подпись на основе данных запроса."""
        if not self.is_authorized():
            raise NotAuthorized("Api key and api secret is required to private endpoints")

        query_string = dict_to_query_string(payload)
        return generate_hmac_sha256_signature(
            self._api_secret,  # type: ignore[attr-defined]
            query_string,
            "hex",
        )

    async def _make_request(
        self,
        method: RequestMethod,
        url: str,
        *,
        params: dict[str, Any] | None = None,
        signed: bool = False,
    ) -> Any:
        """Выполняет HTTP-запрос к эндпоинтам Mexc API.

        Если signed=True, формируется подпись для приватных endpoint'ов:
            - Если переданы params — подпись добавляется в параметры запроса.
            - Если передан data — подпись добавляется в тело запроса.

        Если signed=False, запрос отправляется как публичный.

        Параметры:
            method (`str`): HTTP метод ("GET", "POST", "DELETE" и т.д.).
            url (`str`): Полный URL эндпоинта Mexc API.
            params (`dict | None`): Query-параметры.
            signed (`bool`): Нужно ли подписывать запрос.

        Возвращает:
            `dict`: Ответ в формате JSON.
        """
        # Фильтруем параметры
        payload = filter_params(params) if params else {}

        # Генериуем подпись, если запрос авторизованый
        if signed:
            # Генерируем подпись
            payload["timestamp"] = int(time.time() * 1000)
            payload["recvWindow"] = self._RECV_WINDOW
            payload["signature"] = self._generate_signature(payload)

        # Формируем заголовки запроса
        headers = self._get_headers(signed=signed)

        return await super()._make_request(
            method=method,
            url=url,
            params=payload,
            headers=headers,
        )

    async def request(self, method: RequestMethod, url: str, params: dict, signed: bool) -> dict:
        """Специальный метод для выполнения запросов на эндпоинты, которые не обернуты в клиенте.

        Параметры:
            method (RequestMethod): Метод запроса (GET, POST, PUT, DELETE).
            url (str): URL эндпоинта.
            params (dict): Параметры запроса.
            signed (bool): Флаг, указывающий, требуется ли подпись запроса.

        Возвращает:
            `dict`: Ответ в формате JSON.
        """
        return await self._make_request(method=method, url=url, params=params, signed=signed)

    # topic: Market
    async def ping(self) -> dict:
        """Проверка соединения с REST API.

        https://www.mexc.com/api-docs/spot-v3/market-data-endpoints#test-connectivity
        """
        return await self._make_request("GET", self._BASE_SPOT_URL + "/api/v3/ping")

    async def server_time(self) -> dict:
        """Получение текущего серверного времени.

        https://www.mexc.com/api-docs/spot-v3/market-data-endpoints#check-server-time
        """
        return await self._make_request("GET", self._BASE_SPOT_URL + "/api/v3/time")

    async def default_symbols(self) -> dict:
        """Получение списка торговых пар по умолчанию.

        https://www.mexc.com/api-docs/spot-v3/market-data-endpoints#api-default-symbol
        """
        return await self._make_request("GET", self._BASE_SPOT_URL + "/api/v3/defaultSymbols")

    async def exchange_info(
        self,
        symbol: str | None = None,
        symbols: list[str] | None = None,
    ) -> dict:
        """Получение торговых правил биржи и информации о символах.

        https://www.mexc.com/api-docs/spot-v3/market-data-endpoints#exchange-information
        """
        params = {
            "symbol": symbol,
            "symbols": symbols,
        }

        return await self._make_request(
            "GET", self._BASE_SPOT_URL + "/api/v3/exchangeInfo", params=params
        )

    async def depth(self, symbol: str, limit: int | None = None) -> dict:
        """Получение стакана цен по торговой паре.

        https://www.mexc.com/api-docs/spot-v3/market-data-endpoints#order-book
        """
        params = {
            "symbol": symbol,
            "limit": limit,
        }

        return await self._make_request("GET", self._BASE_SPOT_URL + "/api/v3/depth", params=params)

    async def trades(self, symbol: str, limit: int | None = None) -> list[dict]:
        """Получение списка последних сделок.

        https://www.mexc.com/api-docs/spot-v3/market-data-endpoints#recent-trades-list
        """
        params = {
            "symbol": symbol,
            "limit": limit,
        }

        return await self._make_request(
            "GET", self._BASE_SPOT_URL + "/api/v3/trades", params=params
        )

    async def agg_trades(
        self,
        symbol: str,
        start_time: int | None = None,
        end_time: int | None = None,
        limit: int | None = None,
    ) -> list[dict]:
        """Получение агрегированных сделок по символу.

        https://www.mexc.com/api-docs/spot-v3/market-data-endpoints#compressedaggregate-trades-list
        """
        params = {
            "symbol": symbol,
            "startTime": start_time,
            "endTime": end_time,
            "limit": limit,
        }

        return await self._make_request(
            "GET", self._BASE_SPOT_URL + "/api/v3/aggTrades", params=params
        )

    async def klines(
        self,
        symbol: str,
        interval: str,
        start_time: int | None = None,
        end_time: int | None = None,
        limit: int | None = None,
    ) -> list[list]:
        """Получение свечных данных по торговой паре.

        https://www.mexc.com/api-docs/spot-v3/market-data-endpoints#klinecandlestick-data
        """
        params = {
            "symbol": symbol,
            "interval": interval,
            "startTime": start_time,
            "endTime": end_time,
            "limit": limit,
        }

        return await self._make_request(
            "GET", self._BASE_SPOT_URL + "/api/v3/klines", params=params
        )

    async def avg_price(self, symbol: str) -> dict:
        """Получение средней цены символа за последние минуты.

        https://www.mexc.com/api-docs/spot-v3/market-data-endpoints#current-average-price
        """
        params = {"symbol": symbol}

        return await self._make_request(
            "GET", self._BASE_SPOT_URL + "/api/v3/avgPrice", params=params
        )

    async def ticker_24hr(self, symbol: str | None = None) -> dict | list[dict]:
        """Получение статистики изменения цены за 24 часа.

        https://www.mexc.com/api-docs/spot-v3/market-data-endpoints#24hr-ticker-price-change-statistics
        """
        params = {"symbol": symbol}

        return await self._make_request(
            "GET", self._BASE_SPOT_URL + "/api/v3/ticker/24hr", params=params
        )

    async def ticker_price(self, symbol: str | None = None) -> dict | list[dict]:
        """Получение текущей цены символа.

        https://www.mexc.com/api-docs/spot-v3/market-data-endpoints#symbol-price-ticker
        """
        params = {"symbol": symbol}

        return await self._make_request(
            "GET", self._BASE_SPOT_URL + "/api/v3/ticker/price", params=params
        )

    async def ticker_book_ticker(self, symbol: str | None = None) -> dict | list[dict]:
        """Получение лучших цен и объемов в стакане.

        https://www.mexc.com/api-docs/spot-v3/market-data-endpoints#symbol-order-book-ticker
        """
        params = {"symbol": symbol}

        return await self._make_request(
            "GET", self._BASE_SPOT_URL + "/api/v3/ticker/bookTicker", params=params
        )

    # topic: Spot Account/Trade
    async def kyc_status(self) -> dict:
        """Получение статуса верификации KYC.

        https://www.mexc.com/api-docs/spot-v3/spot-account-trade#query-kyc-status
        """
        return await self._make_request(
            "GET", self._BASE_SPOT_URL + "/api/v3/kyc/status", signed=True
        )

    async def uid(self) -> dict:
        """Получение UID аккаунта.

        https://www.mexc.com/api-docs/spot-v3/spot-account-trade#query-uid
        """
        return await self._make_request("GET", self._BASE_SPOT_URL + "/api/v3/uid", signed=True)

    async def self_symbols(self) -> dict:
        """Получение списка торговых пар, доступных через API аккаунта.

        https://www.mexc.com/api-docs/spot-v3/spot-account-trade#user-api-default-symbol
        """
        return await self._make_request(
            "GET", self._BASE_SPOT_URL + "/api/v3/selfSymbols", signed=True
        )

    async def test_order(
        self,
        symbol: str,
        side: str,
        type: str,
        quantity: str | None = None,
        quote_order_quantity: str | None = None,
        price: str | None = None,
        new_client_order_id: str | None = None,
        stp_mode: str | None = None,
    ) -> dict:
        """Проверка создания нового ордера без отправки на биржу.

        https://www.mexc.com/api-docs/spot-v3/spot-account-trade#test-new-order
        """
        params = {
            "symbol": symbol,
            "side": side,
            "type": type,
            "quantity": quantity,
            "quoteOrderQty": quote_order_quantity,
            "price": price,
            "newClientOrderId": new_client_order_id,
            "stpMode": stp_mode,
        }

        return await self._make_request(
            "POST", self._BASE_SPOT_URL + "/api/v3/order/test", params=params, signed=True
        )

    async def create_order(
        self,
        symbol: str,
        side: str,
        type: str,
        quantity: str | None = None,
        quote_order_quantity: str | None = None,
        price: str | None = None,
        new_client_order_id: str | None = None,
        stp_mode: str | None = None,
    ) -> dict:
        """Создание нового ордера.

        https://www.mexc.com/api-docs/spot-v3/spot-account-trade#new-order
        """
        params = {
            "symbol": symbol,
            "side": side,
            "type": type,
            "quantity": quantity,
            "quoteOrderQty": quote_order_quantity,
            "price": price,
            "newClientOrderId": new_client_order_id,
            "stpMode": stp_mode,
        }

        return await self._make_request(
            "POST", self._BASE_SPOT_URL + "/api/v3/order", params=params, signed=True
        )

    async def batch_orders(self, batch_orders: list[dict]) -> dict | list[dict]:
        """Создание нескольких ордеров одновременно.

        https://www.mexc.com/api-docs/spot-v3/spot-account-trade#batch-orders
        """
        params = {"batchOrders": batch_orders}

        return await self._make_request(
            "POST", self._BASE_SPOT_URL + "/api/v3/batchOrders", params=params, signed=True
        )

    async def cancel_order(
        self,
        symbol: str,
        order_id: str | None = None,
        orig_client_order_id: str | None = None,
        new_client_order_id: str | None = None,
    ) -> dict:
        """Отмена активного ордера.

        https://www.mexc.com/api-docs/spot-v3/spot-account-trade#cancel-order
        """
        params = {
            "symbol": symbol,
            "orderId": order_id,
            "origClientOrderId": orig_client_order_id,
            "newClientOrderId": new_client_order_id,
        }

        return await self._make_request(
            "DELETE", self._BASE_SPOT_URL + "/api/v3/order", params=params, signed=True
        )

    async def cancel_open_orders(self, symbol: str | list[str]) -> list[dict]:
        """Отмена всех открытых ордеров по символу.

        https://www.mexc.com/api-docs/spot-v3/spot-account-trade#cancel-all-open-orders-on-a-symbol
        """
        params = {"symbol": symbol}

        return await self._make_request(
            "DELETE", self._BASE_SPOT_URL + "/api/v3/openOrders", params=params, signed=True
        )

    async def query_order(
        self,
        symbol: str,
        order_id: str | None = None,
        orig_client_order_id: str | None = None,
    ) -> dict:
        """Получение информации об ордере.

        https://www.mexc.com/api-docs/spot-v3/spot-account-trade#query-order
        """
        params = {
            "symbol": symbol,
            "orderId": order_id,
            "origClientOrderId": orig_client_order_id,
        }

        return await self._make_request(
            "GET", self._BASE_SPOT_URL + "/api/v3/order", params=params, signed=True
        )

    async def open_orders(self, symbol: str | None = None) -> list[dict]:
        """Получение списка открытых ордеров.

        https://www.mexc.com/api-docs/spot-v3/spot-account-trade#current-open-orders
        """
        params = {"symbol": symbol}

        return await self._make_request(
            "GET", self._BASE_SPOT_URL + "/api/v3/openOrders", params=params, signed=True
        )

    async def all_orders(
        self,
        symbol: str,
        start_time: int | None = None,
        end_time: int | None = None,
        limit: int | None = None,
    ) -> list[dict]:
        """Получение списка всех ордеров аккаунта по символу.

        https://www.mexc.com/api-docs/spot-v3/spot-account-trade#all-orders
        """
        params = {
            "symbol": symbol,
            "startTime": start_time,
            "endTime": end_time,
            "limit": limit,
        }

        return await self._make_request(
            "GET", self._BASE_SPOT_URL + "/api/v3/allOrders", params=params, signed=True
        )

    async def account(self) -> dict:
        """Получение информации об аккаунте.

        https://www.mexc.com/api-docs/spot-v3/spot-account-trade#account-information
        """
        return await self._make_request("GET", self._BASE_SPOT_URL + "/api/v3/account", signed=True)

    async def my_trades(
        self,
        symbol: str,
        order_id: str | None = None,
        start_time: int | None = None,
        end_time: int | None = None,
        limit: int | None = None,
    ) -> list[dict]:
        """Получение списка сделок аккаунта по символу.

        https://www.mexc.com/api-docs/spot-v3/spot-account-trade#account-trade-list
        """
        params = {
            "symbol": symbol,
            "orderId": order_id,
            "startTime": start_time,
            "endTime": end_time,
            "limit": limit,
        }

        return await self._make_request(
            "GET", self._BASE_SPOT_URL + "/api/v3/myTrades", params=params, signed=True
        )

    async def enable_mx_deduct(self, mx_deduct_enable: bool) -> dict:
        """Включение или отключение списания комиссий в MX.

        https://www.mexc.com/api-docs/spot-v3/spot-account-trade#enable-mx-deduct
        """
        params = {"mxDeductEnable": mx_deduct_enable}

        return await self._make_request(
            "POST", self._BASE_SPOT_URL + "/api/v3/mxDeduct/enable", params=params, signed=True
        )

    async def mx_deduct_status(self) -> dict:
        """Получение статуса списания комиссий в MX.

        https://www.mexc.com/api-docs/spot-v3/spot-account-trade#query-mx-deduct-status
        """
        return await self._make_request(
            "GET", self._BASE_SPOT_URL + "/api/v3/mxDeduct/enable", signed=True
        )

    async def trade_fee(self, symbol: str) -> dict:
        """Получение комиссий по символу.

        https://www.mexc.com/api-docs/spot-v3/spot-account-trade#query-symbol-commission
        """
        params = {"symbol": symbol}

        return await self._make_request(
            "GET", self._BASE_SPOT_URL + "/api/v3/tradeFee", params=params, signed=True
        )

    async def create_strategy_group(self, trade_group_name: str) -> dict:
        """Создание стратегии STP.

        https://www.mexc.com/api-docs/spot-v3/spot-account-trade#create-stp-strategy-group
        """
        params = {"tradeGroupName": trade_group_name}

        return await self._make_request(
            "POST", self._BASE_SPOT_URL + "/api/v3/strategy/group", params=params, signed=True
        )

    async def strategy_group(self, trade_group_name: str) -> dict:
        """Получение информации о стратегии STP.

        https://www.mexc.com/api-docs/spot-v3/spot-account-trade#query-stp-strategy-group
        """
        params = {"tradeGroupName": trade_group_name}

        return await self._make_request(
            "GET", self._BASE_SPOT_URL + "/api/v3/strategy/group", params=params, signed=True
        )

    async def delete_strategy_group(self, trade_group_id: str) -> dict:
        """Удаление стратегии STP.

        https://www.mexc.com/api-docs/spot-v3/spot-account-trade#delete-stp-strategy-group
        """
        params = {"tradeGroupId": trade_group_id}

        return await self._make_request(
            "DELETE", self._BASE_SPOT_URL + "/api/v3/strategy/group", params=params, signed=True
        )

    async def add_strategy_group_uid(self, uid: str | list[str], trade_group_id: str) -> dict:
        """Добавление UID в стратегию STP.

        https://www.mexc.com/api-docs/spot-v3/spot-account-trade#add-uid-to-stp-strategy-group
        """
        params = {
            "uid": uid,
            "tradeGroupId": trade_group_id,
        }

        return await self._make_request(
            "POST", self._BASE_SPOT_URL + "/api/v3/strategy/group/uid", params=params, signed=True
        )

    async def delete_strategy_group_uid(self, uid: str | list[str], trade_group_id: str) -> dict:
        """Удаление UID из стратегии STP.

        https://www.mexc.com/api-docs/spot-v3/spot-account-trade#delete-uid-to-stp-strategy-group
        """
        params = {
            "uid": uid,
            "tradeGroupId": trade_group_id,
        }

        return await self._make_request(
            "DELETE", self._BASE_SPOT_URL + "/api/v3/strategy/group/uid", params=params, signed=True
        )

    # topic: Websocket User Data Streams

    async def create_listen_key(self) -> dict:
        """Создание listen key для пользовательского вебсокета.

        https://www.mexc.com/api-docs/spot-v3/websocket-user-data-streams#generate-listen-key
        """
        return await self._make_request(
            "POST", self._BASE_SPOT_URL + "/api/v3/userDataStream", signed=True
        )

    async def listen_keys(self) -> dict:
        """Получение списка актуальных listen key.

        https://www.mexc.com/api-docs/spot-v3/websocket-user-data-streams#get-valid-listen-keys
        """
        return await self._make_request(
            "GET", self._BASE_SPOT_URL + "/api/v3/userDataStream", signed=True
        )

    async def renew_listen_key(self, listen_key: str) -> dict:
        """Продление срока действия listen key.

        https://www.mexc.com/api-docs/spot-v3/websocket-user-data-streams#extend-listen-key-validity
        """
        params = {"listenKey": listen_key}

        return await self._make_request(
            "PUT", self._BASE_SPOT_URL + "/api/v3/userDataStream", params=params, signed=True
        )

    async def close_listen_key(self, listen_key: str) -> dict:
        """Закрытие listen key для пользовательского вебсокета.

        https://www.mexc.com/api-docs/spot-v3/websocket-user-data-streams#close-listen-key
        """
        params = {"listenKey": listen_key}

        return await self._make_request(
            "DELETE", self._BASE_SPOT_URL + "/api/v3/userDataStream", params=params, signed=True
        )

    # topic: Futures Market endpoints

    async def futures_server_time(self) -> dict:
        """Получение текущего серверного времени фьючерсного API.

        https://www.mexc.com/api-docs/futures/market-endpoints
        """
        return await self._make_request("GET", self._BASE_FUTURES_URL + "/api/v1/contract/ping")

    async def futures_contract_detail(self, symbol: str | None = None) -> dict:
        """Получение информации о фьючерсных контрактах.

        https://www.mexc.com/api-docs/futures/market-endpoints
        """
        params = {"symbol": symbol}

        return await self._make_request(
            "GET", self._BASE_FUTURES_URL + "/api/v1/contract/detail", params=params
        )

    async def futures_support_currencies(self) -> dict:
        """Получение списка поддерживаемых для перевода валют.

        https://www.mexc.com/api-docs/futures/market-endpoints
        """
        return await self._make_request(
            "GET", self._BASE_FUTURES_URL + "/api/v1/contract/support_currencies"
        )

    async def futures_depth(self, symbol: str, limit: int | None = None) -> dict:
        """Получение данных рыночного стакана по контракту.

        https://www.mexc.com/api-docs/futures/market-endpoints
        """
        params = {"limit": limit}

        return await self._make_request(
            "GET", self._BASE_FUTURES_URL + f"/api/v1/contract/depth/{symbol}", params=params
        )

    async def futures_depth_commits(self, symbol: str, limit: int) -> dict:
        """Получение моментального снимка стакана по контракту.

        https://www.mexc.com/api-docs/futures/market-endpoints
        """
        return await self._make_request(
            "GET",
            self._BASE_FUTURES_URL + f"/api/v1/contract/depth_commits/{symbol}/{limit}",
        )

    async def futures_index_price(self, symbol: str) -> dict:
        """Получение индикативной цены контракта.

        https://www.mexc.com/api-docs/futures/market-endpoints
        """
        return await self._make_request(
            "GET", self._BASE_FUTURES_URL + f"/api/v1/contract/index_price/{symbol}"
        )

    async def futures_fair_price(self, symbol: str) -> dict:
        """Получение справедливой цены контракта.

        https://www.mexc.com/api-docs/futures/market-endpoints
        """
        return await self._make_request(
            "GET", self._BASE_FUTURES_URL + f"/api/v1/contract/fair_price/{symbol}"
        )

    async def futures_funding_rate(self, symbol: str) -> dict:
        """Получение текущей ставки финансирования контракта.

        https://www.mexc.com/api-docs/futures/market-endpoints
        """
        return await self._make_request(
            "GET", self._BASE_FUTURES_URL + f"/api/v1/contract/funding_rate/{symbol}"
        )

    async def futures_kline(
        self,
        symbol: str,
        interval: str | None = None,
        start: int | None = None,
        end: int | None = None,
    ) -> dict:
        """Получение свечных данных по контракту.

        https://www.mexc.com/api-docs/futures/market-endpoints#k-line-data
        """
        params = {
            "interval": interval,
            "start": start,
            "end": end,
        }

        return await self._make_request(
            "GET", self._BASE_FUTURES_URL + f"/api/v1/contract/kline/{symbol}", params=params
        )

    async def futures_index_price_kline(
        self,
        symbol: str,
        interval: str | None = None,
        start: int | None = None,
        end: int | None = None,
    ) -> dict:
        """Получение свечей индикативной цены контракта.

        https://www.mexc.com/api-docs/futures/market-endpoints
        """
        params = {
            "interval": interval,
            "start": start,
            "end": end,
        }

        return await self._make_request(
            "GET",
            self._BASE_FUTURES_URL + f"/api/v1/contract/kline/index_price/{symbol}",
            params=params,
        )

    async def futures_fair_price_kline(
        self,
        symbol: str,
        interval: str | None = None,
        start: int | None = None,
        end: int | None = None,
    ) -> dict:
        """Получение свечей справедливой цены контракта.

        https://www.mexc.com/api-docs/futures/market-endpoints
        """
        params = {
            "interval": interval,
            "start": start,
            "end": end,
        }

        return await self._make_request(
            "GET",
            self._BASE_FUTURES_URL + f"/api/v1/contract/kline/fair_price/{symbol}",
            params=params,
        )

    async def futures_deals(self, symbol: str, limit: int | None = None) -> dict:
        """Получение последних сделок по контракту.

        https://www.mexc.com/api-docs/futures/market-endpoints
        """
        params = {"limit": limit}

        return await self._make_request(
            "GET", self._BASE_FUTURES_URL + f"/api/v1/contract/deals/{symbol}", params=params
        )

    async def futures_ticker(self, symbol: str | None = None) -> dict:
        """Получение текущих параметров тренда по контракту.

        https://www.mexc.com/api-docs/futures/market-endpoints#get-contract-trend-data
        """
        params = {"symbol": symbol}

        return await self._make_request(
            "GET", self._BASE_FUTURES_URL + "/api/v1/contract/ticker", params=params
        )

    async def futures_risk_reverse(self) -> dict:
        """Получение текущих балансов страхового фонда.

        https://www.mexc.com/api-docs/futures/market-endpoints
        """
        return await self._make_request(
            "GET", self._BASE_FUTURES_URL + "/api/v1/contract/risk_reverse"
        )

    async def futures_risk_reverse_history(
        self,
        symbol: str,
        page_num: int,
        page_size: int,
    ) -> dict:
        """Получение истории баланса страхового фонда по контракту.

        https://www.mexc.com/api-docs/futures/market-endpoints
        """
        params = {
            "symbol": symbol,
            "page_num": page_num,
            "page_size": page_size,
        }

        return await self._make_request(
            "GET", self._BASE_FUTURES_URL + "/api/v1/contract/risk_reverse/history", params=params
        )

    async def futures_funding_rate_history(
        self,
        symbol: str,
        page_num: int,
        page_size: int,
    ) -> dict:
        """Получение истории ставок финансирования по контракту.

        https://www.mexc.com/api-docs/futures/market-endpoints
        """
        params = {
            "symbol": symbol,
            "page_num": page_num,
            "page_size": page_size,
        }

        return await self._make_request(
            "GET", self._BASE_FUTURES_URL + "/api/v1/contract/funding_rate/history", params=params
        )

    # topic: Futures Account and trading endpoints

    async def futures_account_assets(self) -> dict:
        """Получение сведений по всем валютам фьючерсного аккаунта.

        https://www.mexc.com/api-docs/futures/account-and-trading-endpoints
        """
        return await self._make_request(
            "GET",
            self._BASE_FUTURES_URL + "/api/v1/private/account/assets",
            signed=True,
        )

    async def futures_account_asset(self, currency: str) -> dict:
        """Получение баланса по одной валюте фьючерсного аккаунта.

        https://www.mexc.com/api-docs/futures/account-and-trading-endpoints
        """
        return await self._make_request(
            "GET",
            self._BASE_FUTURES_URL + f"/api/v1/private/account/asset/{currency}",
            signed=True,
        )



__all__ = ["ExchangeInfo"]


import aiohttp

from unicex._abc import IExchangeInfo
from unicex.types import TickerInfoItem

from .client import Client


class ExchangeInfo(IExchangeInfo):
    """Предзагружает информацию о тикерах для биржи Binance."""

    exchange_name = "Binance"
    """Название биржи, на которой работает класс."""

    @classmethod
    async def _load_spot_exchange_info(cls, session: aiohttp.ClientSession) -> None:
        """Загружает информацию о бирже для спотового рынка."""
        exchange_info = await Client(session).exchange_info()
        tickers_info: dict[str, TickerInfoItem] = {}
        for symbol_info in exchange_info["symbols"]:
            filters = {
                flt["filterType"]: flt
                for flt in symbol_info.get("filters", [])
                if "filterType" in flt
            }
            price_filter = filters["PRICE_FILTER"]
            lot_size_filter = filters["LOT_SIZE"]
            tickers_info[symbol_info["symbol"]] = TickerInfoItem(
                tick_step=float(price_filter["tickSize"]),
                tick_precision=None,
                size_step=float(lot_size_filter["stepSize"]),
                size_precision=None,
                contract_size=1,
            )

        cls._tickers_info = tickers_info

    @classmethod
    async def _load_futures_exchange_info(cls, session: aiohttp.ClientSession) -> None:
        """Загружает информацию о бирже для фьючерсного рынка."""
        exchange_info = await Client(session).futures_exchange_info()
        tickers_info: dict[str, TickerInfoItem] = {}
        for symbol_info in exchange_info["symbols"]:
            filters = {
                flt["filterType"]: flt
                for flt in symbol_info.get("filters", [])
                if "filterType" in flt
            }
            price_filter = filters["PRICE_FILTER"]
            lot_size_filter = filters["LOT_SIZE"]
            tickers_info[symbol_info["symbol"]] = TickerInfoItem(
                tick_step=float(price_filter["tickSize"]),
                tick_precision=None,
                size_step=float(lot_size_filter["stepSize"]),
                size_precision=None,
                contract_size=1,
            )

        cls._futures_tickers_info = tickers_info



__all__ = ["UserWebsocket"]

import asyncio
from collections.abc import Awaitable, Callable
from typing import Any

from loguru import logger as _logger

from unicex._base import Websocket
from unicex.exceptions import NotSupported
from unicex.types import AccountType, LoggerLike

from .client import Client


class UserWebsocket:
    """Пользовательский вебсокет Binance с авто‑продлением listenKey.

    Поддержка типов аккаунта: "SPOT" и "FUTURES" (USDT‑M фьючерсы).
    """

    _BASE_SPOT_URL: str = "wss://stream.binance.com:9443"
    """Базовый URL для вебсокета на спот."""

    _BASE_FUTURES_URL: str = "wss://fstream.binance.com"
    """Базовый URL для вебсокета на фьючерсы."""

    _RENEW_INTERVAL: int = 30 * 60
    """Интервал продления listenKey (сек.)"""

    def __init__(
        self,
        callback: Callable[[Any], Awaitable[None]],
        client: Client,
        type: AccountType,
        logger: LoggerLike | None = None,
        **kwargs: Any,  # Не дадим сломаться, если юзер передал ненужные аргументы
    ) -> None:
        """Инициализирует пользовательский вебсокет для работы с биржей Binance.

        Параметры:
            callback (`Callable`): Асинхронная функция обратного вызова, которая принимает сообщение с вебсокета.
            client (`Client`): Авторизованный клиент Binance.
            type (`str`): Тип аккаунта ("SPOT" | "FUTURES").
            logger (`LoggerLike | None`): Логгер для записи логов.
        """
        self._callback = callback
        self._client = client
        self._type = type

        self._listen_key: str | None = None
        self._ws: Websocket | None = None
        self._keepalive_task: asyncio.Task | None = None

        self._logger = logger or _logger

        self._running = False

    @classmethod
    def _create_ws_url(cls, type: AccountType, listen_key: str) -> str:
        """Создает URL для подключения к WebSocket."""
        if type == "FUTURES":
            return f"{cls._BASE_FUTURES_URL}/ws/{listen_key}"
        if type == "SPOT":
            return f"{cls._BASE_SPOT_URL}/ws/{listen_key}"
        raise NotSupported(f"Account type '{type}' not supported")

    async def start(self) -> None:
        """Запускает пользовательский стрим с автопродлением listenKey."""
        self._running = True
        self._listen_key = await self._create_listen_key()
        await self._start_ws(self._create_ws_url(self._type, self._listen_key))  # type: ignore

        # Фоновое продление ключа прослушивания
        self._keepalive_task = asyncio.create_task(self._keepalive_loop())

    async def stop(self) -> None:
        """Останавливает стрим и закрывает listenKey."""
        self._running = False

        # Останавливаем вебсокет
        try:
            if isinstance(self._ws, Websocket):
                await self._ws.stop()
        except Exception as e:
            self._logger.error(f"Error stopping WebSocket: {e}")

        # Ожидаем завершения фонового продления ключа прослушивания
        if isinstance(self._keepalive_task, asyncio.Task):
            try:
                self._keepalive_task.cancel()
                await self._keepalive_task
            except asyncio.CancelledError:
                pass
            except Exception as e:
                self._logger.error(f"Error stopping keepalive task: {e}")

        # Закрываем ключ прослушивания
        try:
            if self._listen_key:
                await self._close_listen_key(self._listen_key)
        except Exception as e:
            self._logger.error(f"Error closing listenKey: {e}")
        finally:
            self._listen_key = None

        self._logger.info("User websocket stopped")

    async def restart(self) -> None:
        """Перезапускает WebSocket для User Data Stream."""
        await self.stop()
        await self.start()

    async def _start_ws(self, ws_url: str) -> None:
        """Запускает WebSocket для User Data Stream."""
        self._ws = Websocket(callback=self._callback, url=ws_url, no_message_reconnect_timeout=None)
        await self._ws.start()
        self._logger.info(f"User websocket started: ...{ws_url[-5:]}")

    async def _keepalive_loop(self) -> None:
        """Фоновый цикл продления listenKey и восстановления сессии при необходимости."""
        while self._running:
            try:
                if self._type == "FUTURES":
                    response = await self._renew_listen_key()
                    listen_key = response.get("listenKey") if isinstance(response, dict) else None
                    if not listen_key:
                        raise RuntimeError(f"Can not renew listenKey: {response}")

                    if listen_key != self._listen_key:
                        self._logger.info(
                            f"Listen key changed: {self._listen_key} -> {listen_key}. Restarting websocket"
                        )
                        await self.restart()
                        return

                elif self._type == "SPOT":
                    await self._renew_listen_key()

                else:
                    raise NotSupported(f"Account type '{self._type}' not supported")

            except Exception as e:
                self._logger.error(f"Error while keeping alive: {e}")
                await self.restart()
                return

            # Ждём до следующего продления
            for _ in range(self._RENEW_INTERVAL):
                if not self._running:
                    return
                await asyncio.sleep(1)

    async def _create_listen_key(self) -> str:
        """Создает новый listenKey для User Data Stream в зависимости от типа аккаунта."""
        if self._type == "FUTURES":
            resp = await self._client.futures_listen_key()
        elif self._type == "SPOT":
            resp = await self._client.listen_key()
        else:
            raise NotSupported(f"Account type '{self._type}' not supported")

        key = resp.get("listenKey") if isinstance(resp, dict) else None
        if not key:
            raise RuntimeError(f"Can not create listenKey: {resp}")
        return key

    async def _renew_listen_key(self) -> dict:
        """Продлевает listenKey. Возвращает новый ключ, если сервер его выдал."""
        if not isinstance(self._listen_key, str):
            raise RuntimeError("listenKey is not a string")
        if self._type == "FUTURES":
            return await self._client.futures_renew_listen_key()
        elif self._type == "SPOT":
            return await self._client.renew_listen_key(self._listen_key)
        else:
            raise NotSupported(f"Account type '{self._type}' not supported")

    async def _close_listen_key(self, listen_key: str) -> None:
        """Закрывает listenKey."""
        if self._type == "FUTURES":
            await self._client.futures_close_listen_key()
        elif self._type == "SPOT":
            await self._client.close_listen_key(listen_key)
        else:
            raise NotSupported(f"Account type '{self._type}' not supported")



"""Модуль, который предоставляет типы данных для работы с библиотекой."""

__all__ = [
    "TickerDailyDict",
    "TickerDailyItem",
    "KlineDict",
    "TradeDict",
    "AggTradeDict",
    "RequestMethod",
    "LoggerLike",
    "AccountType",
    "OpenInterestDict",
    "OpenInterestItem",
    "TickerInfoItem",
    "TickersInfoDict",
    "LiquidationDict",
]

from logging import Logger as LoggingLogger
from typing import Literal, TypedDict

import loguru

type LoggerLike = LoggingLogger | loguru.Logger
"""Объединение логгеров: loguru._logger.Logger или logging.Logger."""

type RequestMethod = Literal["GET", "POST", "PUT", "DELETE", "PATCH"]
"""Типы методов HTTP запросов."""


class TickerDailyItem(TypedDict):
    """Статистика одного тикера за последние 24 часа."""

    p: float
    """Изменение цены за 24 ч."""

    v: float
    """Объем торгов за 24 ч. в монетах."""

    q: float
    """Объем торгов за 24 ч. в долларах."""


type TickerDailyDict = dict[str, TickerDailyItem]
"""Статистика тикеров за последние 24 часа."""


class KlineDict(TypedDict):
    """Модель свечи."""

    s: str
    """Символ."""

    t: int
    """Время открытия. В миллисекундах."""

    o: float
    """Цена открытия свечи."""

    h: float
    """Верхняя точка свечи."""

    l: float  # noqa
    """Нижняя точка свечи."""

    c: float
    """Цена закрытия свечи."""

    v: float
    """Объем свечи. В монетах."""

    q: float
    """Объем свечи. В долларах."""

    T: int | None
    """Время закрытия. В миллисекундах."""

    x: bool | None
    """Флаг закрыта ли свеча."""


class TradeDict(TypedDict):
    """Модель сделки."""

    t: int
    """Время сделки. В миллисекундах."""

    s: str
    """Символ."""

    S: Literal["BUY", "SELL"]
    """Направление сделки."""

    p: float
    """Цена сделки."""

    v: float
    """Объем сделки. В монетах."""


class AggTradeDict(TradeDict):
    """Модель агрегированной сделки."""

    pass


class OpenInterestItem(TypedDict):
    """Модель одного элемента открытого интереса."""

    t: int
    """Время. В миллисекундах."""

    v: float
    """Открытый интерес. В монетах."""


type OpenInterestDict = dict[str, OpenInterestItem]
"""Модель открытого интереса."""


class LiquidationDict(TypedDict):
    """Модель ликвидации."""

    t: int
    """Время. В миллисекундах."""

    s: str
    """Символ."""

    S: Literal["BUY", "SELL"]
    """Сторона."""

    v: float
    """Объем ликвидации. В монетах."""

    p: float
    """Цена ликвидации."""


type AccountType = Literal["SPOT", "FUTURES"]
"""Тип аккаунта."""


class TickerInfoItem(TypedDict):
    """Информация о размерах тиков, ступеней цены и множителя контракта (если есть) для тикера.

    На некоторых биржах удобнее делать округление через precisions, на некоторых через step,
    потому что иногда встречаются шаги, которые не являются степенью 10. Поэтому обязательно
    должны быть определены tick_precision ИЛИ tick_step, а так же size_precision ИЛИ size_step.
    """

    tick_precision: int | None
    """Количество знаков после запятой для цены."""

    tick_step: float | None
    """Шаг одного деления для цены."""

    size_precision: int | None
    """Количество знаков после запятой для объема."""

    size_step: float | None
    """Шаг одного деления для объема."""

    contract_size: float | None
    """Множитель контракта (если есть)."""


type TickersInfoDict = dict[str, TickerInfoItem]
"""Информация о размерах тиков, ступеней цены и множителя контракта (если есть) для всех тикеров."""



__all__ = ["WebsocketManager"]


import json
import time
from collections.abc import Awaitable, Callable
from typing import Any, Literal

from unicex._base import Websocket

from .client import Client

type CallbackType = Callable[[Any], Awaitable[None]]


class WebsocketManager:
    """Менеджер асинхронных вебсокетов для Gateio."""

    _SPOT_URL = "wss://api.gateio.ws/ws/v4/"
    """Адрес вебсокета для спотового рынка."""

    _FUTURES_URL = "wss://fx-ws.gateio.ws/v4/ws/usdt"
    """Адрес вебсокета для фьючерсного рынка."""

    def __init__(self, client: Client | None = None, **ws_kwargs: Any) -> None:
        """Инициализирует менеджер вебсокетов для Mexc.

        Параметры:
            client (`Client | None`): Клиент для выполнения запросов. Нужен, чтобы открыть приватные вебсокеты.
            ws_kwargs (`dict[str, Any]`): Дополнительные аргументы, которые прокидываются в `Websocket`.
        """
        self.client = client
        self._ws_kwargs = ws_kwargs

    def _build_subscription_message(
        self,
        channel: str,
        payload: list,
        event: Literal["subscribe", "unsubscribe"] = "subscribe",
    ) -> str:
        """Формирует JSON сообщение для подписки."""
        return json.dumps(
            {
                "time": int(time.time()),
                "id": int(time.time() * 1e6),
                "channel": channel,
                "event": event,
                "payload": payload,
            }
        )

    def _create_keepalive_message_callable(
        self,
        type: Literal["ping", "pong"],
        market: Literal["spot", "futures"] = "spot",
    ) -> Callable[[], str]:
        """Функция для генерации ping или pong сообщения в рантайме."""
        return lambda: json.dumps(
            {
                "time": int(time.time()),
                "channel": f"{market}.{type}",
            }
        )

    def _create_websocket(
        self,
        callback: CallbackType,
        subscription_messages: list[str],
    ) -> Websocket:
        """Шорткат для создания вебсокета."""
        return Websocket(
            callback=callback,
            url=self._SPOT_URL,
            subscription_messages=subscription_messages,
            ping_message=self._create_keepalive_message_callable("ping", "spot"),
            pong_message=self._create_keepalive_message_callable("pong", "spot"),
            **self._ws_kwargs,
        )

    def _create_futures_websocket(
        self,
        callback: CallbackType,
        subscription_messages: list[str],
    ) -> Websocket:
        """Шорткат для создания фьючерсного вебсокета."""
        return Websocket(
            callback=callback,
            url=self._FUTURES_URL,
            subscription_messages=subscription_messages,
            ping_message=self._create_keepalive_message_callable("ping", "futures"),
            pong_message=self._create_keepalive_message_callable("pong", "futures"),
            **self._ws_kwargs,
        )

    def tickers(
        self,
        callback: CallbackType,
        symbols: list[str],
    ) -> Websocket:
        """Открывает вебсокет с общений информацией о тикерах.

        https://www.gate.com/docs/developers/apiv4/ws/en/#tickers-channel.


        Параметры:
            callback (`CallbackType`): Асинхронная функция обратного вызова для обработки сообщений.
            symbols (`Sequence[str] | None`): Список символов для мультиплекс‑подключения.

        Возвращает:
            `Websocket`: Объект для управления вебсокет соединением.
        """
        return self._create_websocket(
            callback, [self._build_subscription_message("spot.tickers", symbols)]
        )

    def trades(
        self,
        callback: CallbackType,
        symbols: list[str],
    ) -> Websocket:
        """Открывает вебсокет для получения публичных трейдов.

        https://www.gate.com/docs/developers/apiv4/ws/en/#public-trades-channel

        Канал `spot.trades` отправляет сообщение при каждом новом трейде (только со стороны тейкера).

        Параметры:
            callback (`CallbackType`): Асинхронная функция для обработки входящих сообщений.
            symbols (`list[str]`): Список торговых пар (например, ["BTC_USDT", "ETH_USDT"]).

        Возвращает:
            `Websocket`: Объект для управления вебсокет-соединением.
        """
        return self._create_websocket(
            callback, [self._build_subscription_message("spot.trades", symbols)]
        )

    def candlesticks(
        self,
        callback: CallbackType,
        interval: str,
        symbols: list[str],
    ) -> Websocket:
        """Открывает вебсокет для получения данных свечей (candlesticks).

        https://www.gate.com/docs/developers/apiv4/ws/en/#candlesticks-channel

        Канал `spot.candlesticks` передаёт информацию о свечах каждые 2 секунды.

        Параметры:
            callback (`CallbackType`): Асинхронная функция для обработки входящих сообщений.
            interval (`str`): Интервал свечей (например: "1m", "5m", "1h", "1d").
            symbols (`list[str]`): Список торговых пар (например: ["BTC_USDT", "ETH_USDT"]).

        Возвращает:
            `Websocket`: Объект для управления вебсокет-соединением.
        """
        payloads = [[interval, symbol] for symbol in symbols]
        subscription_messages = [
            self._build_subscription_message("spot.candlesticks", payload) for payload in payloads
        ]
        return self._create_websocket(callback, subscription_messages)

    def book_ticker(
        self,
        callback: CallbackType,
        symbols: list[str],
    ) -> Websocket:
        """Открывает вебсокет для получения лучшего бид/аск (best bid / best ask).

        Документация: Best bid or ask price — канал `spot.book_ticker`.
        Канал обновляется с частотой ~10 мс.

        Параметры:
            callback (`CallbackType`): Асинхронная функция для обработки входящих сообщений.
            symbols (`list[str]`): Список торговых пар (например ["BTC_USDT", "ETH_USDT"]).

        Возвращает:
            `Websocket`: Объект для управления вебсокет-соединением.
        """
        return self._create_websocket(
            callback, [self._build_subscription_message("spot.book_ticker", symbols)]
        )

    def order_book_update(
        self,
        callback: CallbackType,
        symbols: list[str],
        interval: str = "100ms",
    ) -> Websocket:
        """Открывает вебсокет для получения обновлений стакана (изменения уровней ордербука).

        https://www.gate.com/docs/developers/apiv4/ws/en/#changed-order-book-levels

        Канал `spot.order_book_update` передаёт изменения уровней стакана с частотой 20 мс или 100 мс.

        Параметры:
            callback (`CallbackType`): Асинхронная функция для обработки входящих сообщений.
            symbols (`list[str]`): Список торговых пар (например: ["BTC_USDT", "ETH_USDT"]).
            interval (`str`): Частота обновлений — "20ms" (глубина 20) или "100ms" (глубина 100). По умолчанию — "100ms".

        Возвращает:
            `Websocket`: Объект для управления вебсокет-соединением.
        """
        payloads = [[symbol, interval] for symbol in symbols]
        subscription_messages = [
            self._build_subscription_message("spot.order_book_update", payload)
            for payload in payloads
        ]
        return self._create_websocket(callback, subscription_messages)

    def order_book(
        self,
        callback: CallbackType,
        symbols: list[str],
        level: Literal["5", "10", "20", "50", "100"] = "20",
        interval: Literal["100ms", "1000ms"] = "1000ms",
    ) -> Websocket:
        """Открывает вебсокет для получения снапшота ордербука ограниченного уровня.

        https://www.gate.com/docs/developers/apiv4/ws/en/#limited-level-full-order-book-snapshot

        Канал `spot.order_book` передаёт полный снимок стакана (ограниченной глубины)
        с частотой 100 мс или 1000 мс.

        Параметры:
            callback (`CallbackType`): Асинхронная функция для обработки входящих сообщений.
            symbols (`list[str]`): Список торговых пар (например ["BTC_USDT", "ETH_USDT"]).
            level (`Literal["5", "10", "20", "50", "100"]`): Глубина стакана. По умолчанию — "20".
            interval (`Literal["100ms", "1000ms"]`): Частота обновлений. По умолчанию — "1000ms".

        Возвращает:
            `Websocket`: Объект для управления вебсокет-соединением.
        """
        payloads = [[symbol, level, interval] for symbol in symbols]
        subscription_messages = [
            self._build_subscription_message("spot.order_book", payload) for payload in payloads
        ]
        return self._create_websocket(callback, subscription_messages)

    def order_book_v2(
        self,
        callback: CallbackType,
        symbols: list[str],
        level: Literal["400", "50"] = "400",
    ) -> Websocket:
        """Открывает вебсокет для получения обновлений ордербука V2 по спотовым контрактам.

        Канал `spot.obu` передаёт изменения ордербука в формате V2.
        Формат подписки для каждого символа: ob.{symbol}.{level}.
        - Level "400": обновления каждые 100 мс
        - Level "50": обновления каждые 20 мс

        Документация: https://www.gate.com/docs/developers/apiv4/ws/en/#order-book-v2-subscription

        Параметры:
            callback (`CallbackType`): Асинхронная функция для обработки входящих сообщений.
            symbols (`list[str]`): Список спотовых торговых пар (например ["BTC_USDT", "ETH_USDT"]).
            level (`Literal["400", "50"]`): Глубина ордербука. По умолчанию — "400".

        Возвращает:
            `Websocket`: Объект для управления вебсокет-соединением.
        """
        payloads = [f"ob.{symbol}.{level}" for symbol in symbols]
        return self._create_websocket(
            callback, [self._build_subscription_message("spot.obu", payloads)]
        )

    def futures_tickers(
        self,
        callback: CallbackType,
        symbols: list[str],
    ) -> Websocket:
        """Открывает вебсокет для получения информации по фьючерсным контрактам.

        Канал `futures.tickers` передаёт данные о:
        - последней цене
        - максимальной/минимальной цене за день
        - дневном объёме
        - изменении цены за 24 часа

        Документация: https://www.gate.com/docs/developers/futures/ws/en/#tickers-api

        Параметры:
            callback (`CallbackType`): Асинхронная функция для обработки входящих сообщений.
            symbols (`list[str]`): Список фьючерсных контрактов (например ["BTC_USDT", "ETH_USDT"]).

        Возвращает:
            `Websocket`: Объект для управления вебсокет-соединением.
        """
        return self._create_futures_websocket(
            callback, [self._build_subscription_message("futures.tickers", symbols)]
        )

    def futures_trades(
        self,
        callback: CallbackType,
        symbols: list[str],
    ) -> Websocket:
        """Открывает вебсокет для получения информации о сделках по фьючерсным контрактам.

        Канал `futures.trades` отправляет сообщение при каждом новом трейде.
        Содержит детали сделки: цену, объём, время и тип сделки.

        Документация: https://www.gate.com/docs/developers/futures/ws/en/#trades-api

        Параметры:
            callback (`CallbackType`): Асинхронная функция для обработки входящих сообщений.
            symbols (`list[str]`): Список фьючерсных контрактов (например ["BTC_USDT", "ETH_USDT"]).

        Возвращает:
            `Websocket`: Объект для управления вебсокет-соединением.
        """
        return self._create_futures_websocket(
            callback, [self._build_subscription_message("futures.trades", symbols)]
        )

    def futures_book_ticker(
        self,
        callback: CallbackType,
        symbols: list[str],
    ) -> Websocket:
        """Открывает вебсокет для получения лучшего бид/аск (best bid / best ask) по фьючерсным контрактам.

        Канал `futures.book_ticker` передаёт обновления лучших цен бид и аск для указанных контрактов.

        Документация: https://www.gate.com/docs/developers/futures/ws/en/#best-ask-bid-subscription

        Параметры:
            callback (`CallbackType`): Асинхронная функция для обработки входящих сообщений.
            symbols (`list[str]`): Список фьючерсных контрактов (например ["BTC_USDT", "ETH_USDT"]).

        Возвращает:
            `Websocket`: Объект для управления вебсокет-соединением.
        """
        return self._create_futures_websocket(
            callback, [self._build_subscription_message("futures.book_ticker", symbols)]
        )

    def futures_order_book_update(
        self,
        callback: CallbackType,
        symbols: list[str],
        frequency: Literal["20ms", "100ms"] = "100ms",
        level: Literal["20", "50", "100"] | None = None,
    ) -> Websocket:
        """Открывает вебсокет для получения обновлений ордербука фьючерсных контрактов.

        Канал `futures.order_book_update` передаёт изменения уровней стакана с заданной частотой.

        Документация: https://www.gate.com/docs/developers/futures/ws/en/#order-book-update-subscription

        Параметры:
            callback (`CallbackType`): Асинхронная функция для обработки входящих сообщений.
            symbols (`list[str]`): Список фьючерсных контрактов (например ["BTC_USDT", "ETH_USDT"]).
            frequency (`Literal["20ms", "100ms"]`): Частота обновлений. По умолчанию — "100ms".
            level (`Literal["20", "50", "100"] | None`): Опциональная глубина ордербука. Только для частоты 20ms разрешено "20".

        Возвращает:
            `Websocket`: Объект для управления вебсокет-соединением.
        """
        payloads = [
            [symbol, frequency] + ([level] if level is not None else []) for symbol in symbols
        ]
        subscription_messages = [
            self._build_subscription_message("futures.order_book_update", payload)
            for payload in payloads
        ]
        return self._create_futures_websocket(callback, subscription_messages)

    def futures_order_book_v2(
        self,
        callback: CallbackType,
        symbols: list[str],
        level: Literal["400", "50"] = "400",
    ) -> Websocket:
        """Открывает вебсокет для получения обновлений ордербука V2 по фьючерсным контрактам.

        Канал `futures.obu` передаёт изменения ордербука в формате V2.
        Формат подписки для каждого символа: ob.{symbol}.{level}.
        - Level "400": обновления каждые 100 мс
        - Level "50": обновления каждые 20 мс

        Документация: https://www.gate.com/docs/developers/futures/ws/en/#order-book-v2-subscription

        Параметры:
            callback (`CallbackType`): Асинхронная функция для обработки входящих сообщений.
            symbols (`list[str]`): Список фьючерсных контрактов (например ["BTC_USDT", "ETH_USDT"]).
            level (`Literal["400", "50"]`): Глубина ордербука. По умолчанию — "400".

        Возвращает:
            `Websocket`: Объект для управления вебсокет-соединением.
        """
        payloads = [f"ob.{symbol}.{level}" for symbol in symbols]
        return self._create_futures_websocket(
            callback, [self._build_subscription_message("futures.obu", payloads)]
        )

    def futures_order_book(
        self,
        callback: CallbackType,
        symbols: list[str],
        limit: Literal["100", "50", "20", "10", "5", "1"] = "20",
        interval: Literal["0"] = "0",
    ) -> Websocket:
        """Открывает вебсокет для получения снапшота ордербука фьючерсных контрактов.

        Канал `futures.order_book` передаёт полный снимок стакана с указанной глубиной и интервалом.

        Документация: https://www.gate.com/docs/developers/futures/ws/en/#legacy-order-book-subscription

        Параметры:
            callback (`CallbackType`): Асинхронная функция для обработки входящих сообщений.
            symbols (`list[str]`): Список фьючерсных контрактов (например ["BTC_USDT", "ETH_USDT"]).
            limit (`Literal["100", "50", "20", "10", "5", "1"]`): Глубина ордербука. По умолчанию — "20".
            interval (`Literal["0"]`): Интервал обновления. По документации всегда "0".

        Возвращает:
            `Websocket`: Объект для управления вебсокет-соединением.
        """
        payloads = [[symbol, limit, interval] for symbol in symbols]
        subscription_messages = [
            self._build_subscription_message("futures.order_book", payload) for payload in payloads
        ]
        return self._create_futures_websocket(callback, subscription_messages)

    def futures_candlesticks(
        self,
        callback: CallbackType,
        interval: str,
        symbols: list[str],
        price_type: Literal["index", "mark"] | None = None,
    ) -> Websocket:
        """Открывает вебсокет для получения данных свечей (candlesticks) по фьючерсным контрактам.

        Канал `futures.candlesticks` передаёт информацию о свечах с указанным интервалом.
        Для подписки на специальные типы цен можно указать `price_type`:
        - "mark_" для mark price
        - "index_" для index price

        Документация: https://www.gate.com/docs/developers/futures/ws/en/#candlesticks-api

        Параметры:
            callback (`CallbackType`): Асинхронная функция для обработки входящих сообщений.
            interval (`str`): Интервал свечей, например "10s", "1m", "5m", "1h", "1d".
            symbols (`list[str]`): Список фьючерсных контрактов (например ["BTC_USDT", "ETH_USDT"]).
            price_type (`Literal["index", "mark"] | None`): Тип цены для свечей. По умолчанию — обычные контракты.

        Возвращает:
            `Websocket`: Объект для управления вебсокет-соединением.
        """
        if price_type:
            prefix = f"{price_type}_"
            symbols = [f"{prefix}{symbol}" for symbol in symbols]

        payloads = [[interval, symbol] for symbol in symbols]
        subscription_messages = [
            self._build_subscription_message("futures.candlesticks", payload)
            for payload in payloads
        ]
        return self._create_futures_websocket(callback, subscription_messages)

    def futures_public_liquidates(
        self,
        callback: CallbackType,
        symbols: list[str],
    ) -> Websocket:
        """Открывает вебсокет для получения информации о ликвидациях по фьючерсным контрактам.

        Канал `futures.public_liquidates` передаёт данные о ликвидационных ордерах.
        Каждый контракт может присылать до одного сообщения о ликвидации в секунду.
        Для подписки на все контракты можно использовать значение ["!all"] в списке `symbols`.

        Документация: https://www.gate.com/docs/developers/futures/ws/en/#public-liquidates-order-api

        Параметры:
            callback (`CallbackType`): Асинхронная функция для обработки входящих сообщений.
            symbols (`list[str]`): Список фьючерсных контрактов (например ["BTC_USDT", "ETH_USDT"]) или ["!all"] для всех контрактов.

        Возвращает:
            `Websocket`: Объект для управления вебсокет-соединением.
        """
        return self._create_futures_websocket(
            callback, [self._build_subscription_message("futures.public_liquidates", symbols)]
        )

    def futures_contract_stats(
        self,
        callback: CallbackType,
        interval: str,
        symbols: list[str],
    ) -> Websocket:
        """Открывает вебсокет для получения статистики по фьючерсным контрактам.

        Канал `futures.contract_stats` передаёт агрегированную статистику по каждому контракту.
        Интервал определяет период агрегирования данных (например, "1m", "5m", "1h", "1d").

        Документация: https://www.gate.com/docs/developers/futures/ws/en/#contract-stats-api

        Параметры:
            callback (`CallbackType`): Асинхронная функция для обработки входящих сообщений.
            interval (`str`): Интервал статистики, например "1m", "5m", "1h", "1d".
            symbols (`list[str]`): Список фьючерсных контрактов (например ["BTC_USDT", "ETH_USDT"]).

        Возвращает:
            `Websocket`: Объект для управления вебсокет-соединением.
        """
        payloads = [[symbol, interval] for symbol in symbols]
        subscription_messages = [
            self._build_subscription_message("futures.contract_stats", payload)
            for payload in payloads
        ]
        return self._create_futures_websocket(callback, subscription_messages)



__all__ = ["Websocket"]

import asyncio
import time
from collections.abc import Awaitable, Callable
from typing import Any, Protocol

import orjson
import websockets
from loguru import logger as _logger
from websockets.asyncio.client import ClientConnection

from unicex.exceptions import QueueOverflowError
from unicex.types import LoggerLike


class Websocket:
    """Базовый класс асинхронного вебсокета."""

    MAX_QUEUE_SIZE: int = 500
    """Максимальная длина очереди."""

    class _DecoderProtocol(Protocol):
        """Протокол декодирования сообщений."""

        def decode(self, message: Any) -> dict: ...

    class _JsonDecoder:
        """Протокол декодирования сообщений в формате JSON."""

        def decode(self, message: Any) -> dict:
            return orjson.loads(message)

    def __init__(
        self,
        callback: Callable[[Any], Awaitable[None]],
        url: str,
        subscription_messages: list[dict] | list[str] | None = None,
        ping_interval: int | float = 10,
        ping_message: str | Callable | None = None,
        pong_message: str | Callable | None = None,
        no_message_reconnect_timeout: int | float | None = 60,
        reconnect_timeout: int | float | None = 5,
        worker_count: int = 1,
        logger: LoggerLike | None = None,
        decoder: type[_DecoderProtocol] = _JsonDecoder,
        **kwargs: Any,  # Не дадим сломаться, если юзер передал ненужные аргументы
    ) -> None:
        """Инициализация вебсокета.

        Параметры:
            callback (`Callable[[Any], Awaitable[None]]`): Обработчик входящих сообщений.
            url (`str`): URL вебсокета.
            subscription_messages (`list[dict] | list[str] | None`): Сообщения для подписки после подключения.
            ping_interval (`int | float`): Интервал отправки ping, сек.
            ping_message (`str | Callable | None`): Сообщение для ping, или функция генерации ping (если не указано — используется ping‑frame).
            pong_message (`str | Callable | None`): Сообщение для pong, или функция генерации pong (если не указано — используется pong‑frame).
            no_message_reconnect_timeout (`int | float | None`): Таймаут ожидания без сообщений до рестарта, сек.
            reconnect_timeout (`int | float | None`): Пауза перед переподключением, сек.
            worker_count (`int`): Количество рабочих задач для обработки сообщений.
            logger (`LoggerLike | None`): Логгер для записи логов.
            decoder (`IDecoder | None`): Декодер для обработки входящих сообщений.
        """
        self._callback = callback
        self._url = url
        self._subscription_messages = subscription_messages or []
        self._ping_interval = ping_interval
        self._ping_message = ping_message
        self._pong_message = pong_message
        self._no_message_reconnect_timeout = no_message_reconnect_timeout
        self._reconnect_timeout = reconnect_timeout or 0
        self._last_message_time = time.monotonic()
        self._worker_count = worker_count
        self._logger = logger or _logger
        self._decoder = decoder()
        self._tasks: list[asyncio.Task] = []
        self._queue = asyncio.Queue()
        self._running = False

    async def start(self) -> None:
        """Запускает вебсокет и рабочие задачи."""
        # Проверяем что вебсокет еще не запущен
        if self._running:
            raise RuntimeError("Websocket is already running")
        self._running = True

        # Запускаем вебсокет
        try:
            await self._connect()
        except Exception as e:
            self._logger.error(f"Failed to connect to websocket: {e}")
            self._running = False
            raise

    async def stop(self) -> None:
        """Останавливает вебсокет и рабочие задачи."""
        self._running = False
        await self._after_disconnect()

    async def restart(self) -> None:
        """Перезапускает вебсокет."""
        await self.stop()
        await asyncio.sleep(self._reconnect_timeout)
        await self.start()

    async def _connect(self) -> None:
        """Подключается к вебсокету и настраивает соединение."""
        self._logger.debug(f"Establishing connection with {self._url}")
        async for conn in websockets.connect(uri=self._url, **self._generate_ws_kwargs()):
            try:
                self._logger.info(f"Websocket connection was established to {self._url}")
                await self._after_connect(conn)

                # Цикл получения сообщений
                while self._running:
                    message = await conn.recv()
                    await self._handle_message(message)

            except websockets.exceptions.ConnectionClosed as e:
                self._logger.error(f"Websocket connection was closed unexpectedly: {e}")
            except Exception as e:
                self._logger.error(f"Unexpected error in websosocket connection: {e}")
            finally:
                # Делаем реконнект только если вебсокет активен, иначе выходим из итератора
                if self._running:
                    await asyncio.sleep(self._reconnect_timeout)
                    await self._after_disconnect()
                else:
                    return  # Выходим из итератора, если вебсокет уже выключен

    async def _handle_message(self, message: str | bytes) -> None:
        """Обрабатывает входящее сообщение вебсокета."""
        try:
            # Обновленяем время последнего сообщения
            self._last_message_time = time.monotonic()

            # Ложим сообщение в очередь, предварительно его сериализуя
            decoded_message = self._decoder.decode(message)
            await self._queue.put(decoded_message)

            # Проверяем размер очереди сообщений и выбрасываем ошибку, если он превышает максимальный размер
            self._check_queue_size()
        except QueueOverflowError:
            self._logger.error("Message queue is overflow")
        except orjson.JSONDecodeError as e:
            if message in ["ping", "pong"]:
                self._logger.debug(f"Received ping message: {message}")
            else:
                self._logger.error(f"Failed to decode JSON message: {message}, error: {e}")
        except Exception as e:
            self._logger.error(f"Unexpected error: {e}")

    def _check_queue_size(self) -> None:
        """Проверяет размер очереди и выбрасывает ошибку при переполнении."""
        qsize = self._queue.qsize()
        if qsize >= self.MAX_QUEUE_SIZE:
            raise QueueOverflowError(f"Message queue is overflow: {qsize}")

    async def _after_connect(self, conn: ClientConnection) -> None:
        """Вызывается после установки соединения."""
        # Подписываемся на топики
        await self._send_subscribe_messages(conn)

        # Обновленяем время последнего сообщения перед каждым подключением
        self._last_message_time = time.monotonic()

        # Запускам задачу для кастомного пинг сообщения
        if self._ping_message:
            self._tasks.append(asyncio.create_task(self._custom_ping_task(conn)))

        # Запускаем healthcheck
        if self._no_message_reconnect_timeout:
            self._tasks.append(asyncio.create_task(self._healthcheck_task()))

        # Запускаем воркеров
        for _ in range(self._worker_count):
            task = asyncio.create_task(self._worker())
            self._tasks.append(task)

    async def _after_disconnect(self) -> None:
        """Вызывается после отключения от вебсокета."""
        current_task = asyncio.current_task()

        # Останавливаем воркеров, исключая задачу, которая уже выполняет остановку
        tasks_to_wait: list[asyncio.Task] = []
        for task in self._tasks:
            if task is current_task:
                continue

            task.cancel()
            tasks_to_wait.append(task)

        # Дожидаемся завершения задач (в т.ч. воркеров)
        if tasks_to_wait:
            results = await asyncio.gather(*tasks_to_wait, return_exceptions=True)
            for task_result in results:
                if isinstance(task_result, asyncio.CancelledError):
                    continue
                if isinstance(task_result, Exception):
                    self._logger.warning(f"Worker raised during shutdown: {task_result}")

        self._tasks.clear()

        # Очистить очередь уже безопасно, после остановки воркеров
        self._queue = asyncio.Queue()

    async def _send_subscribe_messages(self, conn: ClientConnection) -> None:
        """Отправляет сообщения с подпиской на топики, если нужно."""
        for message in self._subscription_messages:
            await conn.send(message)
            self._logger.debug(f"Sent subscribe message: {message}")

    async def _worker(self) -> None:
        """Обрабатывает сообщения из очереди."""
        while self._running:
            try:
                data = await self._queue.get()  # Получаем сообщение
                await self._callback(data)  # Передаем в callback
                self._queue.task_done()
            except asyncio.exceptions.CancelledError:
                break
            except Exception as e:
                self._logger.error(f"Error({type(e)}) while processing message: {e}")

    def _generate_ws_kwargs(self) -> dict:
        """Генерирует аргументы для запуска вебсокета."""
        ws_kwargs = {}
        if self._ping_interval:
            ws_kwargs["ping_interval"] = self._ping_interval
        return ws_kwargs

    async def _custom_ping_task(self, conn: ClientConnection) -> None:
        """Периодически отправляет пользовательский ping."""
        while self._running and self._ping_message:
            try:
                if isinstance(self._ping_message, Callable):
                    ping_message = self._ping_message()
                else:
                    ping_message = self._ping_message
                await conn.send(ping_message)
                self._logger.debug(f"Sent ping message: {ping_message}")
            except Exception as e:
                self._logger.error(f"Error sending ping: {e}")
                return
            await asyncio.sleep(self._ping_interval)

    async def _healthcheck_task(self) -> None:
        """Следит за таймаутом получения сообщений."""
        if not self._no_message_reconnect_timeout:
            return

        while self._running:
            if time.monotonic() - self._last_message_time > self._no_message_reconnect_timeout:
                self._logger.error("Websocket is not responding, restarting...")
                await self.restart()
                return
            await asyncio.sleep(1)

    def __repr__(self) -> str:
        """Репрезентация вебсокета."""
        return f"<Websocket(url={self._url[:15]}...)>"



__all__ = ["Client"]

import hashlib
import hmac
import json
import time
from typing import Any, Literal

from unicex._base import BaseClient
from unicex.exceptions import NotAuthorized
from unicex.types import RequestMethod
from unicex.utils import dict_to_query_string, filter_params


class Client(BaseClient):
    """Клиент для работы с Gateio API."""

    _BASE_URL: str = "https://api.gateio.ws"
    """Базовый URL для REST API Gate.io."""

    def _prepare_request(
        self,
        *,
        method: RequestMethod,
        endpoint: str,
        signed: bool,
        params: dict[str, Any] | None,
        data: dict[str, Any] | None,
    ) -> tuple[str, dict[str, Any] | None, dict[str, Any] | None, dict[str, str]]:
        """Формирует параметры и заголовки для HTTP-запроса."""
        params = filter_params(params) if params else None
        data = filter_params(data) if data else None
        url = f"{self._BASE_URL}{endpoint}"

        timestamp = str(int(time.time()))
        headers: dict[str, str] = {
            "Accept": "application/json",
            "Content-Type": "application/json",
            "Timestamp": timestamp,
        }
        if self._api_key:  # type: ignore[attr-defined]
            headers["KEY"] = self._api_key  # type: ignore[attr-defined]

        if not signed:
            return url, params, data, headers

        if not self.is_authorized():
            raise NotAuthorized("Api key and api secret is required to private endpoints")

        payload_string = json.dumps(data, separators=(",", ":")) if data else ""
        query_string = dict_to_query_string(params) if params else ""
        hashed_payload = hashlib.sha512(payload_string.encode("utf-8")).hexdigest()
        signature_body = (
            f"{method.upper()}\n{endpoint}\n{query_string}\n{hashed_payload}\n{timestamp}"
        )
        signature = hmac.new(
            self._api_secret.encode("utf-8"),  # type: ignore[attr-defined]
            signature_body.encode("utf-8"),
            hashlib.sha512,
        ).hexdigest()
        headers["SIGN"] = signature
        return url, params, data, headers

    async def _make_request(
        self,
        method: RequestMethod,
        endpoint: str,
        signed: bool = False,
        *,
        params: dict[str, Any] | None = None,
        data: dict[str, Any] | None = None,
    ) -> Any:
        """Выполняет HTTP-запрос к Gate.io API."""
        url, params, data, headers = self._prepare_request(
            method=method,
            endpoint=endpoint,
            signed=signed,
            params=params,
            data=data,
        )
        return await super()._make_request(
            method=method,
            url=url,
            params=params,
            data=data,
            headers=headers,
        )

    async def request(
        self,
        method: RequestMethod,
        endpoint: str,
        params: dict[str, Any] | None,
        data: dict[str, Any] | None,
        signed: bool,
    ) -> dict:
        """Специальный метод для выполнения произвольных REST-запросов.

        Параметры:
            method (`RequestMethod`): HTTP-метод запроса ("GET", "POST" и т.д.).
            endpoint (`str`): Относительный путь эндпоинта Gate.io API.
            params (`dict | None`): Query-параметры запроса.
            data (`dict | None`): Тело запроса.
            signed (`bool`): Нужно ли подписывать запрос.

        Возвращает:
            `dict`: Ответ Gate.io API.
        """
        return await self._make_request(
            method=method,
            endpoint=endpoint,
            params=params,
            data=data,
            signed=signed,
        )

    # topic: Spot

    async def currencies(self) -> dict:
        """Получение информации о всех валютах.

        https://www.gate.com/docs/developers/apiv4/en/#query-all-currency-information
        """
        return await self._make_request("GET", "/api/v4/spot/currencies")

    async def currency(self, currency: str) -> dict:
        """Получение информации о конкретной валюте.

        https://www.gate.com/docs/developers/apiv4/en/#query-single-currency-information
        """
        return await self._make_request("GET", f"/api/v4/spot/currencies/{currency}")

    async def currency_pairs(self) -> dict:
        """Получение списка поддерживаемых торговых пар.

        https://www.gate.com/docs/developers/apiv4/en/#query-all-supported-currency-pairs
        """
        return await self._make_request("GET", "/api/v4/spot/currency_pairs")

    async def currency_pair(self, currency_pair: str) -> dict:
        """Получение информации о конкретной торговой паре.

        https://www.gate.com/docs/developers/apiv4/en/#query-single-currency-pair-details
        """
        return await self._make_request("GET", f"/api/v4/spot/currency_pairs/{currency_pair}")

    async def tickers(
        self,
        currency_pair: str | None = None,
        timezone: str | None = None,
    ) -> dict:
        """Получение информации о тикерах торговых пар.

        https://www.gate.com/docs/developers/apiv4/en/#get-currency-pair-ticker-information
        """
        params = {
            "currency_pair": currency_pair,
            "timezone": timezone,
        }

        return await self._make_request("GET", "/api/v4/spot/tickers", params=params)

    async def order_book(
        self,
        currency_pair: str,
        interval: str | None = None,
        limit: int | None = None,
        with_id: bool | None = None,
    ) -> dict:
        """Получение информации о стакане рынка.

        https://www.gate.com/docs/developers/apiv4/en/#get-market-depth-information
        """
        params = {
            "currency_pair": currency_pair,
            "interval": interval,
            "limit": limit,
            "with_id": with_id,
        }

        return await self._make_request("GET", "/api/v4/spot/order_book", params=params)

    async def trades(
        self,
        currency_pair: str,
        limit: int | None = None,
        last_id: str | None = None,
        reverse: bool | None = None,
        from_time: int | None = None,
        to_time: int | None = None,
        page: int | None = None,
    ) -> dict:
        """Получение списка рыночных сделок.

        https://www.gate.com/docs/developers/apiv4/en/#query-market-transaction-records
        """
        params = {
            "currency_pair": currency_pair,
            "limit": limit,
            "last_id": last_id,
            "reverse": reverse,
            "from": from_time,
            "to": to_time,
            "page": page,
        }

        return await self._make_request("GET", "/api/v4/spot/trades", params=params)

    async def candlesticks(
        self,
        currency_pair: str,
        limit: int | None = None,
        from_time: int | None = None,
        to_time: int | None = None,
        interval: str | None = None,
    ) -> dict:
        """Получение данных свечного графика.

        https://www.gate.com/docs/developers/apiv4/en/#market-k-line-chart
        """
        params = {
            "currency_pair": currency_pair,
            "limit": limit,
            "from": from_time,
            "to": to_time,
            "interval": interval,
        }

        return await self._make_request("GET", "/api/v4/spot/candlesticks", params=params)

    async def fee(self, currency_pair: str | None = None) -> dict:
        """Получение комиссий аккаунта по торговым парам.

        https://www.gate.com/docs/developers/apiv4/en/#query-account-fee-rates
        """
        params = {
            "currency_pair": currency_pair,
        }

        return await self._make_request("GET", "/api/v4/spot/fee", params=params, signed=True)

    async def batch_fee(self, currency_pairs: str) -> dict:
        """Получение комиссий аккаунта по нескольким торговым парам.

        https://www.gate.com/docs/developers/apiv4/en/#batch-query-account-fee-rates
        """
        params = {
            "currency_pairs": currency_pairs,
        }

        return await self._make_request("GET", "/api/v4/spot/batch_fee", params=params, signed=True)

    async def accounts(self, currency: str | None = None) -> dict:
        """Получение балансов спотовых аккаунтов.

        https://www.gate.com/docs/developers/apiv4/en/#list-spot-trading-accounts
        """
        params = {
            "currency": currency,
        }

        return await self._make_request("GET", "/api/v4/spot/accounts", params=params, signed=True)

    async def account_book(
        self,
        currency: str | None = None,
        from_time: int | None = None,
        to_time: int | None = None,
        page: int | None = None,
        limit: int | None = None,
        type: str | None = None,
        code: str | None = None,
    ) -> dict:
        """Получение истории транзакций спотового аккаунта.

        https://www.gate.com/docs/developers/apiv4/en/#query-spot-account-transaction-history
        """
        params = {
            "currency": currency,
            "from": from_time,
            "to": to_time,
            "page": page,
            "limit": limit,
            "type": type,
            "code": code,
        }

        return await self._make_request(
            "GET", "/api/v4/spot/account_book", params=params, signed=True
        )

    async def batch_orders(
        self,
        orders: list[dict[str, Any]],
    ) -> dict:
        """Создание нескольких ордеров за один запрос.

        https://www.gate.com/docs/developers/apiv4/en/#batch-place-orders
        """
        # NOTE: Документация описывает тело запроса как массив ордеров.
        # Клиент Gate.io сериализует данные как объект с ключом "orders".
        data = {
            "orders": orders,
        }

        return await self._make_request("POST", "/api/v4/spot/batch_orders", data=data, signed=True)

    async def open_orders(
        self,
        page: int | None = None,
        limit: int | None = None,
        account: str | None = None,
    ) -> dict:
        """Получение списка всех открытых ордеров.

        https://www.gate.com/docs/developers/apiv4/en/#list-all-open-orders
        """
        params = {
            "page": page,
            "limit": limit,
            "account": account,
        }

        return await self._make_request(
            "GET", "/api/v4/spot/open_orders", params=params, signed=True
        )

    async def cross_liquidate_orders(
        self,
        currency_pair: str,
        amount: str,
        price: str,
        text: str | None = None,
        action_mode: str | None = None,
    ) -> dict:
        """Создание ордера для закрытия позиции при отключённой перекрёстной торговле.

        https://www.gate.com/docs/developers/apiv4/en/#close-position-when-cross-currency-is-disabled
        """
        data = {
            "text": text,
            "currency_pair": currency_pair,
            "amount": amount,
            "price": price,
            "action_mode": action_mode,
        }

        return await self._make_request(
            "POST", "/api/v4/spot/cross_liquidate_orders", data=data, signed=True
        )

    async def create_order(
        self,
        currency_pair: str,
        side: str,
        amount: str,
        text: str | None = None,
        type: str | None = None,
        account: str | None = None,
        price: str | None = None,
        time_in_force: str | None = None,
        iceberg: str | None = None,
        auto_borrow: bool | None = None,
        auto_repay: bool | None = None,
        stp_act: str | None = None,
        fee_discount: str | None = None,
        action_mode: str | None = None,
    ) -> dict:
        """Создание нового ордера.

        https://www.gate.com/docs/developers/apiv4/en/#create-an-order
        """
        data = {
            "text": text,
            "currency_pair": currency_pair,
            "type": type,
            "account": account,
            "side": side,
            "amount": amount,
            "price": price,
            "time_in_force": time_in_force,
            "iceberg": iceberg,
            "auto_borrow": auto_borrow,
            "auto_repay": auto_repay,
            "stp_act": stp_act,
            "fee_discount": fee_discount,
            "action_mode": action_mode,
        }

        return await self._make_request("POST", "/api/v4/spot/orders", data=data, signed=True)

    async def orders(
        self,
        currency_pair: str,
        status: str,
        page: int | None = None,
        limit: int | None = None,
        account: str | None = None,
        from_time: int | None = None,
        to_time: int | None = None,
        side: str | None = None,
    ) -> dict:
        """Получение списка ордеров по статусу.

        https://www.gate.com/docs/developers/apiv4/en/#list-orders
        """
        params = {
            "currency_pair": currency_pair,
            "status": status,
            "page": page,
            "limit": limit,
            "account": account,
            "from": from_time,
            "to": to_time,
            "side": side,
        }

        return await self._make_request("GET", "/api/v4/spot/orders", params=params, signed=True)

    async def cancel_all_orders(
        self,
        currency_pair: str | None = None,
        side: str | None = None,
        account: str | None = None,
        action_mode: str | None = None,
    ) -> dict:
        """Отмена всех открытых ордеров по заданным условиям.

        https://www.gate.com/docs/developers/apiv4/en/#cancel-all-open-orders-in-specified-currency-pair
        """
        params = {
            "currency_pair": currency_pair,
            "side": side,
            "account": account,
            "action_mode": action_mode,
        }

        return await self._make_request("DELETE", "/api/v4/spot/orders", params=params, signed=True)

    async def cancel_batch_orders(
        self,
        orders: list[dict[str, Any]],
    ) -> dict:
        """Отмена нескольких ордеров по списку идентификаторов.

        https://www.gate.com/docs/developers/apiv4/en/#cancel-batch-orders-by-specified-id-list
        """
        # NOTE: Документация описывает тело запроса как массив объектов.
        # В текущей реализации данные оборачиваются в объект с ключом "orders".
        data = {
            "orders": orders,
        }

        return await self._make_request(
            "POST", "/api/v4/spot/cancel_batch_orders", data=data, signed=True
        )

    async def order(
        self,
        order_id: str,
        currency_pair: str | None = None,
        account: str | None = None,
    ) -> dict:
        """Получение информации о конкретном ордере.

        https://www.gate.com/docs/developers/apiv4/en/#query-single-order-details
        """
        params = {
            "currency_pair": currency_pair,
            "account": account,
        }

        return await self._make_request(
            "GET", f"/api/v4/spot/orders/{order_id}", params=params, signed=True
        )

    async def amend_order(
        self,
        order_id: str,
        currency_pair: str | None = None,
        account: str | None = None,
        amount: str | None = None,
        price: str | None = None,
        amend_text: str | None = None,
        action_mode: str | None = None,
    ) -> dict:
        """Изменение параметров ордера.

        https://www.gate.com/docs/developers/apiv4/en/#amend-single-order
        """
        params = {
            "currency_pair": currency_pair,
            "account": account,
        }
        data = {
            "currency_pair": currency_pair,
            "account": account,
            "amount": amount,
            "price": price,
            "amend_text": amend_text,
            "action_mode": action_mode,
        }

        return await self._make_request(
            "PATCH",
            f"/api/v4/spot/orders/{order_id}",
            params=params,
            data=data,
            signed=True,
        )

    async def cancel_order(
        self,
        order_id: str,
        currency_pair: str,
        account: str | None = None,
        action_mode: str | None = None,
    ) -> dict:
        """Отмена конкретного ордера.

        https://www.gate.com/docs/developers/apiv4/en/#cancel-single-order
        """
        params = {
            "currency_pair": currency_pair,
            "account": account,
            "action_mode": action_mode,
        }

        return await self._make_request(
            "DELETE",
            f"/api/v4/spot/orders/{order_id}",
            params=params,
            signed=True,
        )

    async def my_trades(
        self,
        currency_pair: str | None = None,
        limit: int | None = None,
        page: int | None = None,
        order_id: str | None = None,
        account: str | None = None,
        from_time: int | None = None,
        to_time: int | None = None,
    ) -> dict:
        """Получение личной истории сделок.

        https://www.gate.com/docs/developers/apiv4/en/#query-personal-trading-records
        """
        params = {
            "currency_pair": currency_pair,
            "limit": limit,
            "page": page,
            "order_id": order_id,
            "account": account,
            "from": from_time,
            "to": to_time,
        }

        return await self._make_request("GET", "/api/v4/spot/my_trades", params=params, signed=True)

    async def countdown_cancel_all(
        self,
        timeout: int,
        currency_pair: str | None = None,
    ) -> dict:
        """Настройка автоматической отмены ордеров по таймеру.

        https://www.gate.com/docs/developers/apiv4/en/#countdown-cancel-orders
        """
        data = {
            "timeout": timeout,
            "currency_pair": currency_pair,
        }

        return await self._make_request(
            "POST", "/api/v4/spot/countdown_cancel_all", data=data, signed=True
        )

    async def amend_batch_orders(
        self,
        orders: list[dict[str, Any]],
    ) -> dict:
        """Изменение параметров нескольких ордеров.

        https://www.gate.com/docs/developers/apiv4/en/#batch-modification-of-orders
        """
        # NOTE: Формат массива ордеров следует уточнить в зависимости от требований API.
        data = {
            "orders": orders,
        }

        return await self._make_request(
            "POST", "/api/v4/spot/amend_batch_orders", data=data, signed=True
        )

    async def insurance_history(
        self,
        business: str,
        currency: str,
        from_time: int,
        to_time: int,
        page: int | None = None,
        limit: int | None = None,
    ) -> dict:
        """Получение истории страхового фонда спотового рынка.

        https://www.gate.com/docs/developers/apiv4/en/#query-spot-insurance-fund-historical-data
        """
        params = {
            "business": business,
            "currency": currency,
            "from": from_time,
            "to": to_time,
            "page": page,
            "limit": limit,
        }

        return await self._make_request(
            "GET", "/api/v4/spot/insurance_history", params=params, signed=True
        )

    async def create_price_order(self, order: dict[str, Any]) -> dict:
        """Создание отложенного ордера с ценовым триггером.

        https://www.gate.com/docs/developers/apiv4/en/#create-price-triggered-order
        """
        return await self._make_request(
            "POST", "/api/v4/spot/price_orders", data=order, signed=True
        )

    async def price_orders(
        self,
        status: str,
        market: str | None = None,
        account: str | None = None,
        limit: int | None = None,
        offset: int | None = None,
    ) -> dict:
        """Получение списка активных отложенных ордеров.

        https://www.gate.com/docs/developers/apiv4/en/#query-running-auto-order-list
        """
        params = {
            "status": status,
            "market": market,
            "account": account,
            "limit": limit,
            "offset": offset,
        }

        return await self._make_request(
            "GET", "/api/v4/spot/price_orders", params=params, signed=True
        )

    async def cancel_price_orders(
        self,
        market: str | None = None,
        account: str | None = None,
    ) -> dict:
        """Отмена всех отложенных ордеров.

        https://www.gate.com/docs/developers/apiv4/en/#cancel-all-auto-orders
        """
        params = {
            "market": market,
            "account": account,
        }

        return await self._make_request(
            "DELETE", "/api/v4/spot/price_orders", params=params, signed=True
        )

    async def price_order(self, order_id: str) -> dict:
        """Получение информации о конкретном отложенном ордере.

        https://www.gate.com/docs/developers/apiv4/en/#query-single-auto-order-details
        """
        return await self._make_request("GET", f"/api/v4/spot/price_orders/{order_id}", signed=True)

    async def cancel_price_order(self, order_id: str) -> dict:
        """Отмена конкретного отложенного ордера.

        https://www.gate.com/docs/developers/apiv4/en/#cancel-single-auto-order
        """
        return await self._make_request(
            "DELETE", f"/api/v4/spot/price_orders/{order_id}", signed=True
        )

    # topic: Futures

    async def futures_contracts(
        self,
        settle: Literal["usdt", "btc"],
        limit: int | None = None,
        offset: int | None = None,
    ) -> dict:
        """Получение списка фьючерсных контрактов.

        https://www.gate.com/docs/developers/apiv4/en/#query-all-futures-contracts
        """
        params = {
            "limit": limit,
            "offset": offset,
        }

        return await self._make_request("GET", f"/api/v4/futures/{settle}/contracts", params=params)

    async def futures_contract(self, settle: str, contract: str) -> dict:
        """Получение информации о конкретном фьючерсном контракте.

        https://www.gate.com/docs/developers/apiv4/en/#query-single-contract-information
        """
        return await self._make_request("GET", f"/api/v4/futures/{settle}/contracts/{contract}")

    async def futures_order_book(
        self,
        settle: str,
        contract: str,
        interval: str | None = None,
        limit: int | None = None,
        with_id: bool | None = None,
    ) -> dict:
        """Получение информации о стакане фьючерсного рынка.

        https://www.gate.com/docs/developers/apiv4/en/#query-futures-market-depth-information
        """
        params = {
            "contract": contract,
            "interval": interval,
            "limit": limit,
            "with_id": with_id,
        }

        return await self._make_request(
            "GET", f"/api/v4/futures/{settle}/order_book", params=params
        )

    async def futures_trades(
        self,
        settle: str,
        contract: str,
        limit: int | None = None,
        offset: int | None = None,
        last_id: str | None = None,
        from_time: int | None = None,
        to_time: int | None = None,
    ) -> dict:
        """Получение списка рыночных сделок по фьючерсам.

        https://www.gate.com/docs/developers/apiv4/en/#futures-market-transaction-records
        """
        params = {
            "contract": contract,
            "limit": limit,
            "offset": offset,
            "last_id": last_id,
            "from": from_time,
            "to": to_time,
        }

        return await self._make_request("GET", f"/api/v4/futures/{settle}/trades", params=params)

    async def futures_candlesticks(
        self,
        settle: str,
        contract: str,
        from_time: int | None = None,
        to_time: int | None = None,
        limit: int | None = None,
        interval: str | None = None,
        timezone: str | None = None,
    ) -> dict:
        """Получение свечных данных по фьючерсному контракту.

        https://www.gate.com/docs/developers/apiv4/en/#futures-market-k-line-chart
        """
        params = {
            "contract": contract,
            "from": from_time,
            "to": to_time,
            "limit": limit,
            "interval": interval,
            "timezone": timezone,
        }

        return await self._make_request(
            "GET", f"/api/v4/futures/{settle}/candlesticks", params=params
        )

    async def futures_premium_index(
        self,
        settle: str,
        contract: str,
        from_time: int | None = None,
        to_time: int | None = None,
        limit: int | None = None,
        interval: str | None = None,
    ) -> dict:
        """Получение данных премиум-индекса.

        https://www.gate.com/docs/developers/apiv4/en/#premium-index-k-line-chart
        """
        params = {
            "contract": contract,
            "from": from_time,
            "to": to_time,
            "limit": limit,
            "interval": interval,
        }

        return await self._make_request(
            "GET", f"/api/v4/futures/{settle}/premium_index", params=params
        )

    async def futures_tickers(
        self,
        settle: str,
        contract: str | None = None,
    ) -> dict:
        """Получение торговой статистики по фьючерсам.

        https://www.gate.com/docs/developers/apiv4/en/#get-all-futures-trading-statistics
        """
        params = {
            "contract": contract,
        }

        return await self._make_request("GET", f"/api/v4/futures/{settle}/tickers", params=params)

    async def futures_funding_rate(
        self,
        settle: str,
        contract: str,
        limit: int | None = None,
        from_time: int | None = None,
        to_time: int | None = None,
    ) -> dict:
        """Получение истории ставок финансирования.

        https://www.gate.com/docs/developers/apiv4/en/#futures-market-historical-funding-rate
        """
        params = {
            "contract": contract,
            "limit": limit,
            "from": from_time,
            "to": to_time,
        }

        return await self._make_request(
            "GET", f"/api/v4/futures/{settle}/funding_rate", params=params
        )

    async def futures_insurance(
        self,
        settle: str,
        limit: int | None = None,
    ) -> dict:
        """Получение истории страхового фонда фьючерсного рынка.

        https://www.gate.com/docs/developers/apiv4/en/#futures-market-insurance-fund-history
        """
        params = {
            "limit": limit,
        }

        return await self._make_request("GET", f"/api/v4/futures/{settle}/insurance", params=params)

    async def futures_contract_stats(
        self,
        settle: str,
        contract: str,
        from_time: int | None = None,
        interval: str | None = None,
        limit: int | None = None,
    ) -> dict:
        """Получение статистики по фьючерсному контракту.

        https://www.gate.com/docs/developers/apiv4/en/#futures-statistics
        """
        params = {
            "contract": contract,
            "from": from_time,
            "interval": interval,
            "limit": limit,
        }

        return await self._make_request(
            "GET", f"/api/v4/futures/{settle}/contract_stats", params=params
        )

    async def futures_index_constituents(self, settle: str, index: str) -> dict:
        """Получение состава индекса.

        https://www.gate.com/docs/developers/apiv4/en/#query-index-constituents
        """
        return await self._make_request(
            "GET", f"/api/v4/futures/{settle}/index_constituents/{index}"
        )

    async def futures_liq_orders(
        self,
        settle: str,
        contract: str | None = None,
        from_time: int | None = None,
        to_time: int | None = None,
        limit: int | None = None,
    ) -> dict:
        """Получение истории ликвидаций.

        https://www.gate.com/docs/developers/apiv4/en/#query-liquidation-order-history
        """
        params = {
            "contract": contract,
            "from": from_time,
            "to": to_time,
            "limit": limit,
        }

        return await self._make_request(
            "GET", f"/api/v4/futures/{settle}/liq_orders", params=params
        )

    async def futures_risk_limit_tiers(
        self,
        settle: str,
        contract: str | None = None,
        limit: int | None = None,
        offset: int | None = None,
    ) -> dict:
        """Получение лимитов риска по фьючерсным контрактам.

        https://www.gate.com/docs/developers/apiv4/en/#query-risk-limit-tiers
        """
        params = {
            "contract": contract,
            "limit": limit,
            "offset": offset,
        }

        return await self._make_request(
            "GET", f"/api/v4/futures/{settle}/risk_limit_tiers", params=params
        )

    async def futures_accounts(self, settle: str) -> dict:
        """Получение информации о фьючерсном аккаунте.

        https://www.gate.com/docs/developers/apiv4/en/#get-futures-account
        """
        return await self._make_request("GET", f"/api/v4/futures/{settle}/accounts", signed=True)

    async def futures_account_book(
        self,
        settle: str,
        contract: str | None = None,
        limit: int | None = None,
        offset: int | None = None,
        from_time: int | None = None,
        to_time: int | None = None,
        type_: str | None = None,
    ) -> dict:
        """Получение истории изменений фьючерсного аккаунта.

        https://www.gate.com/docs/developers/apiv4/en/#query-futures-account-change-history
        """
        params = {
            "contract": contract,
            "limit": limit,
            "offset": offset,
            "from": from_time,
            "to": to_time,
            "type": type_,
        }

        return await self._make_request(
            "GET", f"/api/v4/futures/{settle}/account_book", params=params, signed=True
        )

    async def futures_positions(
        self,
        settle: str,
        holding: bool | None = None,
        limit: int | None = None,
        offset: int | None = None,
    ) -> dict:
        """Получение списка позиций пользователя.

        https://www.gate.com/docs/developers/apiv4/en/#get-user-position-list
        """
        params = {
            "holding": holding,
            "limit": limit,
            "offset": offset,
        }

        return await self._make_request(
            "GET", f"/api/v4/futures/{settle}/positions", params=params, signed=True
        )

    async def futures_position(
        self,
        settle: str,
        contract: str,
    ) -> dict:
        """Получение информации о позиции по контракту.

        https://www.gate.com/docs/developers/apiv4/en/#get-single-position-information
        """
        return await self._make_request(
            "GET", f"/api/v4/futures/{settle}/positions/{contract}", signed=True
        )

    async def futures_update_margin(
        self,
        settle: str,
        contract: str,
        change: str,
    ) -> dict:
        """Изменение маржи позиции.

        https://www.gate.com/docs/developers/apiv4/en/#update-position-margin
        """
        params = {
            "change": change,
        }

        return await self._make_request(
            "POST",
            f"/api/v4/futures/{settle}/positions/{contract}/margin",
            params=params,
            signed=True,
        )

    async def futures_update_leverage(
        self,
        settle: str,
        contract: str,
        leverage: str,
        cross_leverage_limit: str | None = None,
        pid: int | None = None,
    ) -> dict:
        """Изменение кредита позиции.

        https://www.gate.com/docs/developers/apiv4/en/#update-position-leverage
        """
        params = {
            "leverage": leverage,
            "cross_leverage_limit": cross_leverage_limit,
            "pid": pid,
        }

        return await self._make_request(
            "POST",
            f"/api/v4/futures/{settle}/positions/{contract}/leverage",
            params=params,
            signed=True,
        )

    async def futures_switch_cross_mode(
        self,
        settle: str,
        mode: str,
        contract: str,
    ) -> dict:
        """Переключение режима маржи позиции.

        https://www.gate.com/docs/developers/apiv4/en/#switch-position-margin-mode
        """
        data = {
            "mode": mode,
            "contract": contract,
        }

        return await self._make_request(
            "POST", f"/api/v4/futures/{settle}/positions/cross_mode", data=data, signed=True
        )

    async def futures_dual_comp_switch_cross_mode(
        self,
        settle: str,
        mode: str,
        contract: str,
    ) -> dict:
        """Переключение режима маржи в хедж-режиме.

        https://www.gate.com/docs/developers/apiv4/en/#switch-between-cross-and-isolated-margin-modes-under-hedge-mode
        """
        data = {
            "mode": mode,
            "contract": contract,
        }

        return await self._make_request(
            "POST",
            f"/api/v4/futures/{settle}/dual_comp/positions/cross_mode",
            data=data,
            signed=True,
        )

    async def futures_update_risk_limit(
        self,
        settle: str,
        contract: str,
        risk_limit: str,
    ) -> dict:
        """Изменение лимита риска позиции.

        https://www.gate.com/docs/developers/apiv4/en/#update-position-risk-limit
        """
        params = {
            "risk_limit": risk_limit,
        }

        return await self._make_request(
            "POST",
            f"/api/v4/futures/{settle}/positions/{contract}/risk_limit",
            params=params,
            signed=True,
        )

    async def futures_set_dual_mode(
        self,
        settle: str,
        dual_mode: bool,
    ) -> dict:
        """Настройка режима двойных позиций.

        https://www.gate.com/docs/developers/apiv4/en/#set-position-mode
        """
        params = {
            "dual_mode": dual_mode,
        }

        return await self._make_request(
            "POST", f"/api/v4/futures/{settle}/dual_mode", params=params, signed=True
        )

    async def futures_dual_comp_position(
        self,
        settle: str,
        contract: str,
    ) -> dict:
        """Получение информации о позиции в двойном режиме.

        https://www.gate.com/docs/developers/apiv4/en/#get-position-information-in-dual-mode
        """
        return await self._make_request(
            "GET", f"/api/v4/futures/{settle}/dual_comp/positions/{contract}", signed=True
        )

    async def futures_dual_comp_update_margin(
        self,
        settle: str,
        contract: str,
        change: str,
        dual_side: str,
    ) -> dict:
        """Изменение маржи позиции в двойном режиме.

        https://www.gate.com/docs/developers/apiv4/en/#update-position-margin-in-dual-mode
        """
        params = {
            "change": change,
            "dual_side": dual_side,
        }

        return await self._make_request(
            "POST",
            f"/api/v4/futures/{settle}/dual_comp/positions/{contract}/margin",
            params=params,
            signed=True,
        )

    async def futures_dual_comp_update_leverage(
        self,
        settle: str,
        contract: str,
        leverage: str,
        cross_leverage_limit: str | None = None,
    ) -> dict:
        """Изменение плеча позиции в двойном режиме.

        https://www.gate.com/docs/developers/apiv4/en/#update-position-leverage-in-dual-mode
        """
        params = {
            "leverage": leverage,
            "cross_leverage_limit": cross_leverage_limit,
        }

        return await self._make_request(
            "POST",
            f"/api/v4/futures/{settle}/dual_comp/positions/{contract}/leverage",
            params=params,
            signed=True,
        )

    async def futures_dual_comp_update_risk_limit(
        self,
        settle: str,
        contract: str,
        risk_limit: str,
    ) -> dict:
        """Изменение лимита риска в двойном режиме.

        https://www.gate.com/docs/developers/apiv4/en/#update-position-risk-limit-in-dual-mode
        """
        params = {
            "risk_limit": risk_limit,
        }

        return await self._make_request(
            "POST",
            f"/api/v4/futures/{settle}/dual_comp/positions/{contract}/risk_limit",
            params=params,
            signed=True,
        )

    async def futures_create_order(self, settle: str, order: dict[str, Any]) -> dict:
        """Создание фьючерсного ордера.

        https://www.gate.com/docs/developers/apiv4/en/#place-futures-order
        """
        return await self._make_request(
            "POST", f"/api/v4/futures/{settle}/orders", data=order, signed=True
        )

    async def futures_orders(
        self,
        settle: str,
        status: str,
        contract: str | None = None,
        limit: int | None = None,
        offset: int | None = None,
        last_id: str | None = None,
    ) -> dict:
        """Получение списка фьючерсных ордеров по статусу.

        https://www.gate.com/docs/developers/apiv4/en/#query-futures-order-list
        """
        params = {
            "contract": contract,
            "status": status,
            "limit": limit,
            "offset": offset,
            "last_id": last_id,
        }

        return await self._make_request(
            "GET", f"/api/v4/futures/{settle}/orders", params=params, signed=True
        )

    async def futures_cancel_all_orders(
        self,
        settle: str,
        contract: str,
        side: str | None = None,
        exclude_reduce_only: bool | None = None,
        text: str | None = None,
    ) -> dict:
        """Отмена всех открытых фьючерсных ордеров по контракту.

        https://www.gate.com/docs/developers/apiv4/en/#cancel-all-orders-with-open-status
        """
        params = {
            "contract": contract,
            "side": side,
            "exclude_reduce_only": exclude_reduce_only,
            "text": text,
        }

        return await self._make_request(
            "DELETE", f"/api/v4/futures/{settle}/orders", params=params, signed=True
        )

    async def futures_orders_timerange(
        self,
        settle: str,
        contract: str | None = None,
        from_time: int | None = None,
        to_time: int | None = None,
        limit: int | None = None,
        offset: int | None = None,
    ) -> dict:
        """Получение списка фьючерсных ордеров за период.

        https://www.gate.com/docs/developers/apiv4/en/#query-futures-order-list-by-time-range
        """
        params = {
            "contract": contract,
            "from": from_time,
            "to": to_time,
            "limit": limit,
            "offset": offset,
        }

        return await self._make_request(
            "GET", f"/api/v4/futures/{settle}/orders_timerange", params=params, signed=True
        )

    async def futures_create_orders_batch(
        self,
        settle: str,
        orders: list[dict[str, Any]],
    ) -> dict:
        """Создание нескольких фьючерсных ордеров за один запрос.

        https://www.gate.com/docs/developers/apiv4/en/#place-batch-futures-orders
        """
        # NOTE: Документация требует массив объектов ордеров, запрос сериализуется как JSON-массив.
        return await self._make_request(
            "POST",
            f"/api/v4/futures/{settle}/batch_orders",
            data=orders,  # type: ignore
            signed=True,
        )

    async def futures_order(
        self,
        settle: str,
        order_id: str,
    ) -> dict:
        """Получение информации о фьючерсном ордере.

        https://www.gate.com/docs/developers/apiv4/en/#query-single-order-details
        """
        return await self._make_request(
            "GET", f"/api/v4/futures/{settle}/orders/{order_id}", signed=True
        )

    async def futures_cancel_order(
        self,
        settle: str,
        order_id: str,
    ) -> dict:
        """Отмена конкретного фьючерсного ордера.

        https://www.gate.com/docs/developers/apiv4/en/#cancel-single-order
        """
        return await self._make_request(
            "DELETE", f"/api/v4/futures/{settle}/orders/{order_id}", signed=True
        )

    async def futures_amend_order(
        self,
        settle: str,
        order_id: str,
        size: int | None = None,
        price: str | None = None,
        amend_text: str | None = None,
        text: str | None = None,
    ) -> dict:
        """Изменение параметров фьючерсного ордера.

        https://www.gate.com/docs/developers/apiv4/en/#amend-single-order
        """
        data = {
            "size": size,
            "price": price,
            "amend_text": amend_text,
            "text": text,
        }

        return await self._make_request(
            "PUT",
            f"/api/v4/futures/{settle}/orders/{order_id}",
            data=data,
            signed=True,
        )

    async def futures_my_trades(
        self,
        settle: str,
        contract: str | None = None,
        order: int | None = None,
        limit: int | None = None,
        offset: int | None = None,
        last_id: str | None = None,
    ) -> dict:
        """Получение личной истории сделок по фьючерсам.

        https://www.gate.com/docs/developers/apiv4/en/#query-personal-trading-records
        """
        params = {
            "contract": contract,
            "order": order,
            "limit": limit,
            "offset": offset,
            "last_id": last_id,
        }

        return await self._make_request(
            "GET", f"/api/v4/futures/{settle}/my_trades", params=params, signed=True
        )

    async def futures_my_trades_timerange(
        self,
        settle: str,
        contract: str | None = None,
        from_time: int | None = None,
        to_time: int | None = None,
        limit: int | None = None,
        offset: int | None = None,
        role: str | None = None,
    ) -> dict:
        """Получение личной истории сделок за период.

        https://www.gate.com/docs/developers/apiv4/en/#query-personal-trading-records-by-time-range
        """
        params = {
            "contract": contract,
            "from": from_time,
            "to": to_time,
            "limit": limit,
            "offset": offset,
            "role": role,
        }

        return await self._make_request(
            "GET", f"/api/v4/futures/{settle}/my_trades_timerange", params=params, signed=True
        )

    async def futures_position_close(
        self,
        settle: str,
        contract: str | None = None,
        limit: int | None = None,
        offset: int | None = None,
        from_time: int | None = None,
        to_time: int | None = None,
        side: str | None = None,
        pnl: str | None = None,
    ) -> dict:
        """Получение истории закрытых позиций.

        https://www.gate.com/docs/developers/apiv4/en/#query-position-close-history
        """
        params = {
            "contract": contract,
            "limit": limit,
            "offset": offset,
            "from": from_time,
            "to": to_time,
            "side": side,
            "pnl": pnl,
        }

        return await self._make_request(
            "GET", f"/api/v4/futures/{settle}/position_close", params=params, signed=True
        )

    async def futures_liquidates(
        self,
        settle: str,
        contract: str | None = None,
        limit: int | None = None,
        offset: int | None = None,
        from_time: int | None = None,
        to_time: int | None = None,
        at: int | None = None,
    ) -> dict:
        """Получение истории ликвидаций аккаунта.

        https://www.gate.com/docs/developers/apiv4/en/#query-liquidation-history
        """
        params = {
            "contract": contract,
            "limit": limit,
            "offset": offset,
            "from": from_time,
            "to": to_time,
            "at": at,
        }

        return await self._make_request(
            "GET", f"/api/v4/futures/{settle}/liquidates", params=params, signed=True
        )

    async def futures_auto_deleverages(
        self,
        settle: str,
        contract: str | None = None,
        limit: int | None = None,
        offset: int | None = None,
        from_time: int | None = None,
        to_time: int | None = None,
        at: int | None = None,
    ) -> dict:
        """Получение истории автоматического снижения плеча (ADL).

        https://www.gate.com/docs/developers/apiv4/en/#query-adl-auto-deleveraging-order-information
        """
        params = {
            "contract": contract,
            "limit": limit,
            "offset": offset,
            "from": from_time,
            "to": to_time,
            "at": at,
        }

        return await self._make_request(
            "GET", f"/api/v4/futures/{settle}/auto_deleverages", params=params, signed=True
        )

    async def futures_countdown_cancel_all(
        self,
        settle: str,
        timeout: int,
        contract: str | None = None,
    ) -> dict:
        """Настройка таймера автоматической отмены фьючерсных ордеров.

        https://www.gate.com/docs/developers/apiv4/en/#countdown-cancel-orders
        """
        data = {
            "timeout": timeout,
            "contract": contract,
        }

        return await self._make_request(
            "POST", f"/api/v4/futures/{settle}/countdown_cancel_all", data=data, signed=True
        )

    async def futures_fee(
        self,
        settle: str,
        contract: str | None = None,
    ) -> dict:
        """Получение ставок комиссий на фьючерсном рынке.

        https://www.gate.com/docs/developers/apiv4/en/#query-futures-market-trading-fee-rates
        """
        params = {
            "contract": contract,
        }

        return await self._make_request(
            "GET", f"/api/v4/futures/{settle}/fee", params=params, signed=True
        )

    async def futures_cancel_orders_batch(
        self,
        settle: str,
        order_ids: list[str],
    ) -> dict:
        """Отмена списка фьючерсных ордеров по идентификаторам.

        https://www.gate.com/docs/developers/apiv4/en/#cancel-batch-orders-by-specified-id-list
        """
        return await self._make_request(
            "POST",
            f"/api/v4/futures/{settle}/batch_cancel_orders",
            data=order_ids,  # type: ignore
            signed=True,
        )

    async def futures_amend_orders_batch(
        self,
        settle: str,
        orders: list[dict[str, Any]],
    ) -> dict:
        """Изменение параметров нескольких фьючерсных ордеров.

        https://www.gate.com/docs/developers/apiv4/en/#batch-modify-orders-by-specified-ids
        """
        return await self._make_request(
            "POST",
            f"/api/v4/futures/{settle}/batch_amend_orders",
            data=orders,  # type: ignore
            signed=True,
        )

    async def futures_risk_limit_table(
        self,
        settle: str,
        table_id: str,
    ) -> dict:
        """Получение таблицы лимитов риска по идентификатору.

        https://www.gate.com/docs/developers/apiv4/en/#query-risk-limit-table-by-table-id
        """
        params = {
            "table_id": table_id,
        }

        return await self._make_request(
            "GET", f"/api/v4/futures/{settle}/risk_limit_table", params=params
        )

    async def futures_create_price_order(
        self,
        settle: str,
        order: dict[str, Any],
    ) -> dict:
        """Создание фьючерсного ордера с ценовым триггером.

        https://www.gate.com/docs/developers/apiv4/en/#create-price-triggered-order
        """
        return await self._make_request(
            "POST", f"/api/v4/futures/{settle}/price_orders", data=order, signed=True
        )

    async def futures_price_orders(
        self,
        settle: str,
        status: str,
        contract: str | None = None,
        limit: int | None = None,
        offset: int | None = None,
    ) -> dict:
        """Получение списка отложенных ордеров по фьючерсам.

        https://www.gate.com/docs/developers/apiv4/en/#query-auto-order-list
        """
        params = {
            "status": status,
            "contract": contract,
            "limit": limit,
            "offset": offset,
        }

        return await self._make_request(
            "GET", f"/api/v4/futures/{settle}/price_orders", params=params, signed=True
        )

    async def futures_cancel_price_orders(
        self,
        settle: str,
        contract: str | None = None,
    ) -> dict:
        """Отмена всех отложенных фьючерсных ордеров.

        https://www.gate.com/docs/developers/apiv4/en/#cancel-all-auto-orders
        """
        params = {
            "contract": contract,
        }

        return await self._make_request(
            "DELETE", f"/api/v4/futures/{settle}/price_orders", params=params, signed=True
        )

    async def futures_price_order(
        self,
        settle: str,
        order_id: str,
    ) -> dict:
        """Получение информации о ценовом триггерном ордере.

        https://www.gate.com/docs/developers/apiv4/en/#query-single-auto-order-details
        """
        return await self._make_request(
            "GET", f"/api/v4/futures/{settle}/price_orders/{order_id}", signed=True
        )

    async def futures_cancel_price_order(
        self,
        settle: str,
        order_id: str,
    ) -> dict:
        """Отмена ценового триггерного ордера.

        https://www.gate.com/docs/developers/apiv4/en/#cancel-single-auto-order
        """
        return await self._make_request(
            "DELETE", f"/api/v4/futures/{settle}/price_orders/{order_id}", signed=True
        )
